# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .cfg_base import TestConfigParserBase 
from pierky.arouteserver.config.bogons import ConfigParserBogons


class TestConfigParserBogons(TestConfigParserBase):

    FILE_PATH = "config.d/bogons.yml"
    CONFIG_PARSER_CLASS = ConfigParserBogons
    SHORT_DESCR = "Bogons config parser"

    def test_valid_cfg(self):
        """{}: valid configuration"""
        self._contains_err()

    def test_unknown_statement(self):
        """{}: unknown statement"""
        self.cfg[0]["test"] = 1
        self._contains_err("Error in bogon definition: Unknown statement 'test' in prefix list entry definition.")

    def test_missing_statement(self):
        """{}: missing statement"""
        del self.cfg[0]["prefix"]
        self._contains_err("Error in bogon definition: Missing 'prefix' in prefix list entry.")
        self.cfg[0]["prefix"] = "192.168.0.0"
        self._contains_err()

    def test_invalid_ipv4_id(self):
        """{}: invalid IPv4 prefix ID"""
        self.cfg[0]["prefix"] = "1000.0.0.1"
        self._contains_err("Error in bogon definition: Invalid prefix ID: 1000.0.0.1.")

    def test_invalid_len_ipv4(self):
        """{}: invalid IPv4 prefix len"""
        self.cfg[0]["prefix"] = "192.0.2.0"
        for l in (-1, 33):
            self.cfg[0]["length"] = l
            self._contains_err("Invalid prefix length: {}".format(l))

    def test_invalid_ipv6_id(self):
        """{}: invalid IPv6 prefix ID"""
        self.cfg[0]["prefix"] = "fe80::1Z"
        self._contains_err("Invalid prefix ID: fe80::1Z")

    def test_invalid_len_ipv6(self):
        """{}: invalid IPv6 prefix len"""
        self.cfg[0]["prefix"] = "fe80::1"
        for l in (-1, 129):
            self.cfg[0]["length"] = l
            self._contains_err("Invalid prefix length: {}".format(l))



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
try:
    import mock
except ImportError:
    import unittest.mock as mock
import os

from pierky.arouteserver.arin_db_dump import ARINWhoisDBDump
from pierky.arouteserver.registro_br_db_dump import RegistroBRWhoisDBDump
from pierky.arouteserver.cached_objects import CachedObject
from pierky.arouteserver.config.validators import ValidatorPrefixListEntry
from pierky.arouteserver.enrichers.irrdb import IRRDBConfigEnricher_ASNs, \
                                                IRRDBConfigEnricher_Prefixes
from pierky.arouteserver.enrichers.rtt import RTTGetterConfigEnricher
from pierky.arouteserver.ipaddresses import IPNetwork
from pierky.arouteserver.irrdb import ASSet, RSet
from pierky.arouteserver.peering_db import PeeringDBInfo, PeeringDBNet, \
                                           PeeringDBNetNeverViaRouteServers
from pierky.arouteserver.ripe_rpki_cache import RIPE_RPKI_ROAs


class MockedEnv(object):

    def load(mocked_env, subdir, filename, ret_type=None):
        rel_path = os.path.join(subdir, filename)

        if rel_path in mocked_env.mocked_files:
            return mocked_env.mocked_files[rel_path]

        if mocked_env.base_dir is None:
            raise ValueError("base_dir is missing")

        path = os.path.join(mocked_env.base_dir, subdir, filename)
        with open(path, "r") as f:
            if ret_type == "json":
                return json.load(f)
            return f.read()

    def do_mock_rttgetter(mocked_env):

        def _mock_RTTGetter(self):
            rtts = {}
            for k in mocked_env.base_inst.RTT:
                if k in mocked_env.base_inst.DATA:
                    # it is a prefix ID and not an IP address
                    rtts[mocked_env.base_inst.DATA[k]] = \
                        mocked_env.base_inst.RTT[k]
                else:
                    rtts[k] = mocked_env.base_inst.RTT[k]

            for client in self.builder.cfg_clients.cfg["clients"]:
                if client["ip"] in rtts:
                    client["rtt"] = rtts[client["ip"]]

        mock_RTTGetter = mock.patch.object(
            RTTGetterConfigEnricher, "enrich", autospec=True
        ).start()
        mock_RTTGetter.side_effect = _mock_RTTGetter

    def do_mock_irr(mocked_env):

        def _mock_ASSet_run_cmd(self, cmd):
            asn_list = set()
            for obj_name in self.object_names:
                raw = mocked_env.load("irrdb_data",
                                      "asset_{}.json".format(obj_name),
                                      ret_type="json")
                asn_list.update(raw["asn_list"])
            return json.dumps({"asn_list": list(asn_list)}).encode()

        def _mock_RSet_run_cmd(self, cmd):
            prefix_list = []
            for obj_name in self.object_names:
                raw = mocked_env.load("irrdb_data",
                                      "rset_{}_ipv{}.json".format(obj_name,
                                                                  self.ip_ver),
                                      ret_type="json")
                for prefix in raw["prefix_list"]:
                    if prefix not in prefix_list:
                        prefix_list.append(prefix)
            return json.dumps({"prefix_list": prefix_list}).encode()

        mock_ASSet_run_cmd = mock.patch.object(
            ASSet, "_run_cmd", autospec=True
        ).start()
        mock_ASSet_run_cmd.side_effect = _mock_ASSet_run_cmd

        mock_RSet_run_cmd = mock.patch.object(
            RSet, "_run_cmd", autospec=True
        ).start()
        mock_RSet_run_cmd.side_effect = _mock_RSet_run_cmd

    def do_mock_irrdb(mocked_env):

        def add_prefix_to_list(prefix_name, allow_longer_prefixes, lst):
            obj = IPNetwork(mocked_env.base_inst.DATA[prefix_name])
            lst.append(
                ValidatorPrefixListEntry().validate({
                    "prefix": obj.ip,
                    "length": obj.prefixlen,
                    "comment": prefix_name,
                    "exact": not allow_longer_prefixes
                })
            )

        def _mock_ASSet(self):
            self.prepare()
            for as_set_bundle_id in self.builder.irrdb_info:
                record = self.builder.irrdb_info[as_set_bundle_id]
                asns = []
                for as_set_name in record.object_names:
                    if as_set_name not in mocked_env.base_inst.AS_SET:
                        continue
                    asns.extend(mocked_env.base_inst.AS_SET[as_set_name])
                if asns:
                    record.save("asns", asns)

        def _mock_RSet(self):
            self.prepare()
            allow_longer_prefixes = self.builder.cfg_general["filtering"]["irrdb"]["allow_longer_prefixes"]
            for as_set_bundle_id in self.builder.irrdb_info:
                record = self.builder.irrdb_info[as_set_bundle_id]
                prefixes = []
                for as_set_name in record.object_names:
                    if as_set_name not in mocked_env.base_inst.R_SET:
                        continue
                    for prefix_name in mocked_env.base_inst.R_SET[as_set_name]:
                        add_prefix_to_list(prefix_name, allow_longer_prefixes,
                                           prefixes)
                if prefixes:
                    record.save("prefixes", prefixes)

        mock_ASSet = mock.patch.object(
            IRRDBConfigEnricher_ASNs, "enrich", autospec=True
        ).start()
        mock_ASSet.side_effect = _mock_ASSet

        mock_RSet = mock.patch.object(
            IRRDBConfigEnricher_Prefixes, "enrich", autospec=True
        ).start()
        mock_RSet.side_effect = _mock_RSet

    def do_mock_cached_objects(mocked_env):

        def load_data_from_cache(self):
            return False

        def save_data_to_cache(self):
            return

        mock_load_data_from_cache = mock.patch.object(
            CachedObject, "load_data_from_cache", autospec=True
        ).start()
        mock_load_data_from_cache.side_effect = load_data_from_cache

        mock_save_data_to_cache = mock.patch.object(
            CachedObject, "save_data_to_cache", autospec=True
        ).start()
        mock_save_data_to_cache.side_effect = save_data_to_cache

    def do_mock_peering_db(mocked_env):

        def get_data_from_peeringdb(self):
            return mocked_env.load(
                "peeringdb_data",
                self._get_peeringdb_url(),
                ret_type="json"
            )

        def get_url_net(self):
            return "net_{}.json".format(self.asn)

        def get_url_never_via_route_servers(self):
            return "never_via_route_servers.json"

        mock_get_data_from_peeringdb = mock.patch.object(
            PeeringDBInfo, "_get_data_from_peeringdb", autospec=True
        ).start()
        mock_get_data_from_peeringdb.side_effect = get_data_from_peeringdb

        mock_get_url_net = mock.patch.object(
            PeeringDBNet, "_get_peeringdb_url", autospec=True
        ).start()
        mock_get_url_net.side_effect = get_url_net

        mock_get_url_net = mock.patch.object(
            PeeringDBNetNeverViaRouteServers, "_get_peeringdb_url", autospec=True
        ).start()
        mock_get_url_net.side_effect = get_url_never_via_route_servers

    def do_mock_ripe_rpki_cache(mocked_env):

        def get_data(self):
            return mocked_env.load("ripe-rpki-cache", "ripe-rpki-cache.json",
                                   ret_type="json")

        mock_get_data = mock.patch.object(
            RIPE_RPKI_ROAs, "_get_data", autospec=True
        ).start()
        mock_get_data.side_effect = get_data

    def do_mock_arin_db_dump(mocked_env):

        def get_data(self):
            return mocked_env.load("arin_whois_db", "dump.json",
                                   ret_type="json")

        mock_get_data = mock.patch.object(
            ARINWhoisDBDump, "_get_data", autospec=True
        ).start()
        mock_get_data.side_effect = get_data

    def do_mock_registrobr_db_dump(mocked_env):

        def get_data(self):
            return mocked_env.load("registrobr_whois_db", "dump.txt")

        mock_get_data = mock.patch.object(
            RegistroBRWhoisDBDump, "_get_data", autospec=True
        ).start()
        mock_get_data.side_effect = get_data

    def __init__(mocked_env, base_inst=None, base_dir=None, default=True,
                 **kwargs):
        """
        - bypass_cache:

          Mock CachedObject.load_data_from_cache() and .save_data_to_cache()
          methods to bypass cache.
          This is automatically set to True when other "cache users"
          functions are mocked.

        - peering_db:

          Mock the PeeringDBInfo._get_data_from_peeringdb() and
          PeeringDBNet._get_peeringdb_url() methods.

          It reads data from the <base_dir>/peeringdb_data/net_<ASN>.json

          Implies bypass_cache.

        - ripe_rpki_cache:

          Mock the RIPE_RPKI_ROAs._get_data() method.

          The content of the RIPE RPKI cache is read from the
          <base_dir>/ripe-rpki-cache/ripe-rpki-cache.json local path.

          Implies bypass_cache.

        - irr:

          Mock ASSet and RSet _run_cmd() methods.

          Raw data as returned by bgpq3 are read from the local files
          <base_dir>/irrdb_data/asset_<name>.json and
          <base_dir>/irrdb_data/rset_<name>_ipv[4|6].json
          Sets of unique ASNs and prefixes are built and returned in the
          bgpq3 format.

          Implies bypass_cache.

        - irrdb:

          Mock IRRDBConfigEnricher_ASNs and IRRDBConfigEnricher_Prefixes
          enrich() method.

          It uses base_inst.DATA, base_inst.AS_SET and base_inst.R_SET to
          read ASNs and prefixes and to save them into IRRDB records:
          record.save("asns") / record.save("prefixes").

          Implies bypass_cache.

        - rttgetter:

          Mock the RTTGetterConfigEnricher.enrich() method.

          It uses base_inst.DATA and base_inst.RTT to fill clients' rtt
          attribute.

        - arin_db_dump:

          Mock the ARINWhoisDBDump._get_data() method.

          The content of the ARIN DB dump is read from the local
          <base_dir>/arin_whois_db/dump.json file.

        - registrobr_db_dump:

          Mock the RegistroBRWhoisDBDump._get_data() method.

          The content of the Registro.br DB dump is read from the local
          <base_dir>/registrobr_whois_db/dump.json file.
        """
        mocked_env.base_inst = base_inst

        mocked_env.bypass_cache = kwargs.get("bypass_cache", default)

        mocked_env.peering_db = kwargs.get("peering_db", default)

        mocked_env.ripe_rpki_cache = kwargs.get("ripe_rpki_cache", default)

        mocked_env.irr = kwargs.get("irr", default)

        mocked_env.irrdb = kwargs.get("irrdb",
                                      default and base_inst is not None)

        if mocked_env.irrdb:
            assert base_inst is not None, \
                "when irrdb is True, base_inst is needed"

        mocked_env.rttgetter = kwargs.get("rttgetter",
                                          default and base_inst is not None)

        if mocked_env.rttgetter:
            assert base_inst is not None, \
                "when rttgetter is True, base_inst is needed"

        mocked_env.arin_db_dump = kwargs.get("arin_db_dump", default)

        mocked_env.registrobr_db_dump = kwargs.get("registrobr_db_dump", default)

        mocked_env.base_dir = base_dir

        # Setup mocks

        if mocked_env.peering_db or mocked_env.ripe_rpki_cache or \
            mocked_env.irr or mocked_env.irrdb or mocked_env.arin_db_dump or \
            mocked_env.registrobr_db_dump:
            mocked_env.bypass_cache = True

        if mocked_env.bypass_cache:
            mocked_env.do_mock_cached_objects()

        if mocked_env.peering_db:
            mocked_env.do_mock_peering_db()

        if mocked_env.ripe_rpki_cache:
            mocked_env.do_mock_ripe_rpki_cache()

        if mocked_env.irr:
            mocked_env.do_mock_irr()

        if mocked_env.irrdb:
            mocked_env.do_mock_irrdb()

        if mocked_env.rttgetter:
            mocked_env.do_mock_rttgetter()

        if mocked_env.arin_db_dump:
            mocked_env.do_mock_arin_db_dump()

        if mocked_env.registrobr_db_dump:
            mocked_env.do_mock_registrobr_db_dump()

        mocked_env.mocked_files = {}

    @staticmethod
    def stopall():
        mock.patch.stopall()



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import yaml

from pierky.arouteserver.config.validators import ValidatorPrefixListEntry
from pierky.arouteserver.errors import ConfigError


class TestPrefixList(unittest.TestCase):

    def _exp_err(self, exp_err=None):
        if not exp_err:
            self._exp_ok()
            return
        try:
            ValidatorPrefixListEntry().validate(self.v)
        except ConfigError as e:
            if not exp_err in str(e):
                self.fail("Expected '{}', received '{}'".format(exp_err, str(e)))

    def _exp_ok(self):
        try:
            ValidatorPrefixListEntry().validate(self.v)
        except ConfigError as e:
            self.fail("Expected success, received '{}'".format(str(e)))

    def _match(self, prefix, len, exact, ge, le, comment):
        ValidatorPrefixListEntry().validate(self.v)
        self.assertEqual(self.v["prefix"], prefix)
        self.assertEqual(self.v["length"], len)
        self.assertEqual(self.v["exact"], exact)
        self.assertEqual(self.v["le"], le)
        self.assertEqual(self.v["ge"], ge)
        self.assertEqual(self.v["comment"], comment)

    def test_bad_prefixes(self):
        """Prefix list parser: bad prefix list entries"""
        self.v = {
            "a": 1
        }
        self._exp_err("Unknown statement 'a' in prefix list entry definition")

        self.v = {
            "prefix": "192.168.0.0"
        }
        self._exp_err("Missing 'length' in prefix list entry")

        self.v["length"] = 33
        self._exp_err("Invalid prefix length: 33")

        self.v = {
            "prefix": "192.168.0.0/24",
            "length": 24
        }
        self._exp_err("Invalid prefix ID: 192.168.0.0/24")

        self.v = {
            "prefix": "2001:db8::",
            "length": -1
        }
        self._exp_err("Invalid prefix length: -1")
        self.v["length"] = 129
        self._exp_err("Invalid prefix length: 129")

    def test_bad_prefix_le_ge(self):
        """Prefix list parser: bad 'le' and 'ge'"""
        self.v = {
            "prefix": "192.168.0.0",
            "length": 16
        }
        self.v["ge"] = 15
        self._exp_err("'ge' (15) must be greater than or equal to the prefix-len (16)")

        self.v["ge"] = 24
        self.v["le"] = 33
        self._exp_err("'le' (33) must be less than or equal to the max prefix-len (32)")

        self.v["ge"] = 30
        self.v["le"] = 24
        self._exp_err("'ge' must be less than or equal to 'le'")

        self.v["ge"] = 24
        self.v["le"] = 30
        self.v["exact"] = True
        self._exp_err("Can't set 'ge' and 'le' when 'exact' is True")

    def test_valid_prefixes(self):
        """Prefix list parser: valid prefix list entries"""

        self.v = {
            "prefix": "192.168.0.0",
            "length": 16
        }
        self._match("192.168.0.0", 16, False, None, None, None)

        self.v["comment"] = "test"
        self._match("192.168.0.0", 16, False, None, None, "test")

        self.v["exact"] = True
        self._match("192.168.0.0", 16, True, None, None, "test")
        
        self.v["exact"] = False

        self.v["ge"] = "24"
        self._match("192.168.0.0", 16, False, 24, None, "test")

        self.v["le"] = "24"
        self._match("192.168.0.0", 16, False, 24, 24, "test")

        self.v["le"] = "30"
        self._match("192.168.0.0", 16, False, 24, 30, "test")

        self.v["ge"] = 16
        self.v["le"] = 16
        self._match("192.168.0.0", 16, False, 16, 16, "test")

        self.v["ge"] = 32
        self.v["le"] = 32
        self._match("192.168.0.0", 16, False, 32, 32, "test")



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six
import unittest

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.ipaddresses import IPNetwork
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class BasicScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    AS_SET = {
        "AS-AS1": [1],
        "AS-AS1_CUSTOMERS": [101, 103, 104],
        "AS-AS2": [2],
        "AS-AS2_CUSTOMERS": [101, 103],
    }
    R_SET = {
        "AS-AS1": [
            "AS1_allowed_prefixes",
            "pref_len1"
        ],
        "AS-AS1_CUSTOMERS": [
            "AS101_allowed_prefixes",
            "AS103_allowed_prefixes",
        ],
        "AS-AS2": [
            "AS2_allowed_prefixes"
        ],
        "AS-AS2_CUSTOMERS": [
            "AS101_allowed_prefixes",
            "AS103_allowed_prefixes",
        ],
    }
    RTT = {
        "AS1_1_IPAddress": 0.1,
        "AS1_2_IPAddress": 5,
        "AS2_1_IPAddress": 17.3,
        "AS3_1_IPAddress": 123.8,
        "AS4_1_IPAddress": 600
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            # Run AS3 as soon as possible because it's configured with passive
            # session, so while other instance come up rs has more time to
            # setup the session with it.
            cls.CLIENT_INSTANCE_CLASS(
                "AS3",
                cls.DATA["AS3_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS3.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1_1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_2",
                cls.DATA["AS1_2_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1_2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS4",
                cls.DATA["AS4_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS4.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS101",
                cls.DATA["AS101_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS101.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            )
        ]

    def set_instance_variables(self):
        self.AS1_1 = self._get_instance_by_name("AS1_1")
        self.AS1_2 = self._get_instance_by_name("AS1_2")
        self.AS2 = self._get_instance_by_name("AS2")
        # AS3 is passive, rs client configured with passive: False
        self.AS3 = self._get_instance_by_name("AS3")
        self.AS4 = self._get_instance_by_name("AS4")
        self.AS101 = self._get_instance_by_name("AS101")
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1_1)
        self.session_is_up(self.rs, self.AS1_2)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS3)
        self.session_is_up(self.rs, self.AS4)
        self.session_is_up(self.AS101, self.AS1_1)
        self.session_is_up(self.AS101, self.AS1_2)
        self.session_is_up(self.AS101, self.AS2)

    def test_021_session_configured_via_local_files(self):
        """{}: session configured via local include files"""

        # A dummy session is configured using local include files.
        # The following tests if those files are really included.
        self.session_exists(self.rs, self.DATA["RoutesCollector_IPAddress"])

    def test_030_good_prefixes_received_by_rs(self):
        """{}: good prefixes received by rs"""
        if isinstance(self.rs, BIRDInstance):
            ext_comm_rpki_unknown = ["generic:0x43000000:0x1"]
        else:
            ext_comm_rpki_unknown = []
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS1_1,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_2,
                           next_hop=self.AS1_2, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS1_2,
                           next_hop=self.AS1_2, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        # AS1_good3 is announced by AS1_2 with NEXT_HOP = AS1_1
        self.receive_route(self.rs, self.DATA["AS1_good3"], self.AS1_2,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2,
                           next_hop=self.AS2, as_path="2",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS2_good2"], self.AS2,
                           next_hop=self.AS2, as_path="2",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])

        # rs should not receive prefixes with the following criteria
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS2)
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS2)
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS1_1)
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS2_good2"], self.AS1_1)

        # AS_PATH should match the expectations
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1")
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2")

        # NEXT_HOP should match the expectations
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2", next_hop=self.AS2)

    def test_030_good_prefixes_because_of_rpki_roas_as_route_objects_1(self):
        """{}: good prefixes because of use_rpki_roas_as_route_objects: exact"""

        prefix = self.DATA["AS101_roa_routeobj_1"]
        self.receive_route(self.rs, prefix,
                           self.AS1_1, as_path="1 101", next_hop=self.AS1_1)
        self.receive_route(self.rs, prefix,
                           self.AS2, as_path="2 101", next_hop=self.AS2)

    def test_030_good_prefixes_because_of_rpki_roas_as_route_objects_2(self):
        """{}: good prefixes because of use_rpki_roas_as_route_objects: covering"""

        prefix = self.DATA["AS101_roa_routeobj_3"]
        self.receive_route(self.rs, prefix,
                           self.AS1_1, as_path="1 101", next_hop=self.AS1_1)
        self.receive_route(self.rs, prefix,
                           self.AS2, as_path="2 101", next_hop=self.AS2)

    def test_030_good_prefixes_because_of_arin_db_1(self):
        """{}: good prefixes because of use_arin_bulk_whois_data"""

        prefix = self.DATA["AS104_arin_1"]
        self.receive_route(self.rs, prefix,
                           self.AS1_1, as_path="1 101 104",
                           next_hop=self.AS1_1)

        # AS101 peers with AS2 too but that route should be reject
        # because only AS1 has AS104 in its AS-SET
        self.receive_route(self.rs, prefix,
                           self.AS2, as_path="2 101 104", next_hop=self.AS2,
                           filtered=True, reject_reason=9)

        self.receive_route(self.AS4, prefix, self.rs, as_path="1 101 104",
                           next_hop=self.AS1_1)

    def test_030_good_prefixes_because_of_registrobr_db_1(self):
        """{}: good prefixes because of use_registrobr_bulk_whois_data"""

        prefix = self.DATA["AS104_nicbr_1"]
        self.receive_route(self.rs, prefix,
                           self.AS1_1, as_path="1 101 104",
                           next_hop=self.AS1_1)

        # AS101 peers with AS2 too but that route should be reject
        # because only AS1 has AS104 in its AS-SET
        self.receive_route(self.rs, prefix,
                           self.AS2, as_path="2 101 104", next_hop=self.AS2,
                           filtered=True, reject_reason=9)

        self.receive_route(self.AS4, prefix, self.rs, as_path="1 101 104",
                           next_hop=self.AS1_1)

    def test_030_good_prefixes_received_by_rs_nonclient_nexthop(self):
        """{}: good prefixes received by rs: non-client NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS2_nonclient_nexthop1"],
                           as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_030_good_prefixes_because_of_irrdb_whitelist(self):
        """{}: good prefixes received by rs: IRRdb white-list"""

        self.receive_route(self.rs, self.DATA["AS1_whitel_1"], as_path="1 1011")
        self.receive_route(self.rs, self.DATA["AS1_whitel_4"], as_path="1 1011")
        self.receive_route(self.rs, self.DATA["AS1_whitel_5"], as_path="1 1000")

    def test_040_bad_prefixes_received_by_rs_bogon(self):
        """{}: bad prefixes received by rs: bogon"""
        self.receive_route(self.rs, self.DATA["bogon1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=2)
        self.log_contains(self.rs, "prefix is bogon - REJECTING " + self.DATA["bogon1"])

    def test_040_bad_prefixes_received_by_rs_global_blacklist(self):
        """{}: bad prefixes received by rs: global blacklist"""

        self.receive_route(self.rs, self.DATA["local1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=3)
        self.log_contains(self.rs, "prefix is in global blacklist - REJECTING " + self.DATA["local1"])

    def test_040_bad_prefixes_received_by_rs_prefix_len(self):
        """{}: bad prefixes received by rs: invalid prefix-len"""

        ip_ver = IPNetwork(self.DATA["pref_len1"]).version

        self.receive_route(self.rs, self.DATA["pref_len1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=13)
        pref_len = 7 if ip_ver == 4 else 16
        pref_len_range = "8-24" if ip_ver == 4 else "17-48"
        self.log_contains(self.rs, "prefix len [{}] not in {} - REJECTING {}".format(
            pref_len, pref_len_range, self.DATA["pref_len1"]))

    def test_040_bad_prefixes_received_by_rs_leftmost_asn(self):
        """{}: bad prefixes received by rs: left-most ASN"""

        self.receive_route(self.rs, self.DATA["peer_as1"], self.AS1_1,
                           as_path="2 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=6)
        self.log_contains(self.rs, "invalid left-most ASN [2] - REJECTING " + self.DATA["peer_as1"])

    def test_040_bad_prefixes_received_by_rs_invalid_aspath(self):
        """{}: bad prefixes received by rs: invalid ASN in AS-PATH"""

        self.receive_route(self.rs, self.DATA["invalid_asn1"], self.AS1_1,
                           as_path="1 65536 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=7)
        self.log_contains(self.rs, "AS_PATH [(path 1 65536 1)] contains invalid ASN - REJECTING " + self.DATA["invalid_asn1"])

    def test_040_bad_prefixes_received_by_rs_transitfree_as_path(self):
        """{}: bad prefixes received by rs: transit-free ASN in AS-PATH"""

        self.receive_route(self.rs, self.DATA["AS101_transitfree_1"],
                           self.AS1_1, as_path="1 101 174",
                           next_hop=self.AS1_1, filtered=True, reject_reason=8)
        self.log_contains(self.rs, "AS_PATH [(path 1 101 174)] contains transit-free ASN - REJECTING " + self.DATA["AS101_transitfree_1"])

    def test_040_bad_prefixes_received_by_rs_transitfree_as_path_transit_peer(self):
        """{}: bad prefixes received by rs: transit-free ASN in AS-PATH from a transit peer"""

        self.receive_route(self.rs, self.DATA["AS3_transitfree_2"],
                           self.AS3, as_path="3 174 33",
                           next_hop=self.AS3, filtered=True, reject_reason=8)
        self.log_contains(self.rs, "AS_PATH [(path 3 174 33)] contains transit-free ASN - REJECTING " + self.DATA["AS3_transitfree_2"])

    def test_040_bad_prefixes_received_by_rs_never_via_rs_peeringdb(self):
        """{}: bad prefixes received by rs: never via route servers ASN in AS-PATH (PeeringDB)"""

        self.receive_route(self.rs, self.DATA["AS101_neverviars_1"],
                           self.AS1_1, as_path="1 101 666",
                           next_hop=self.AS1_1, filtered=True, reject_reason=15)
        self.log_contains(self.rs, "AS_PATH [(path 1 101 666)] contains never via route-servers ASN - REJECTING " + self.DATA["AS101_neverviars_1"])

    def test_040_bad_prefixes_received_by_rs_never_via_rs_asn(self):
        """{}: bad prefixes received by rs: never via route servers ASN in AS-PATH (asns list)"""

        self.receive_route(self.rs, self.DATA["AS101_neverviars_2"],
                           self.AS1_1, as_path="1 101 777",
                           next_hop=self.AS1_1, filtered=True, reject_reason=15)
        self.log_contains(self.rs, "AS_PATH [(path 1 101 777)] contains never via route-servers ASN - REJECTING " + self.DATA["AS101_neverviars_2"])

    def test_040_bad_prefixes_received_by_rs_aspath_len(self):
        """{}: bad prefixes received by rs: AS_PATH len"""

        self.receive_route(self.rs, self.DATA["aspath_len1"], self.AS1_1,
                           as_path="1 2 2 2 2 2 2 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=1)
        self.log_contains(self.rs, "AS_PATH len [8] longer than 6 - REJECTING " + self.DATA["aspath_len1"])

    def test_040_bad_prefixes_received_by_rs_client_blacklist(self):
        """{}: bad prefixes received by rs: client blacklist"""

        self.receive_route(self.rs, self.DATA["AS3_blacklist1"], self.AS3,
                           as_path="3", next_hop=self.AS3,
                           filtered=True, reject_reason=11)
        self.log_contains(self.rs, "prefix is in client's blacklist - REJECTING " + self.DATA["AS3_blacklist1"])

    def test_040_bad_prefixes_received_by_rs_invalid_nexthop(self):
        """{}: bad prefixes received by rs: invalid NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_2,
                           as_path="1 101", next_hop=self.AS101,
                           filtered=True, reject_reason=5)
        self.log_contains(self.rs, "NEXT_HOP [" + self.AS101.ip + "] not allowed - REJECTING " + self.DATA["AS101_good1"])

    def test_040_bad_prefixes_received_by_rs_unknown_nonclient_nexthop(self):
        """{}: bad prefixes received by rs: unknown NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS2_nonclient_nexthop2"],
                           self.AS2, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop2_nh"],
                           filtered=True, reject_reason=5)
        self.log_contains(self.rs, "NEXT_HOP [" + self.DATA["AS2_nonclient_nexthop2_nh"] + "] not allowed - REJECTING " + self.DATA["AS2_nonclient_nexthop2"])

    def test_040_bad_prefixes_received_by_rs_no_rset(self):
        """{}: bad prefixes received by rs: prefix not in AS-SET"""

        # AS101_no_rset is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route(self.rs, self.DATA["AS101_no_rset"], self.AS1_1,
                           as_path="1 101", next_hop=self.AS1_1,
                           filtered=True, reject_reason=12)
        self.receive_route(self.rs, self.DATA["AS101_no_rset"], self.AS2,
                           as_path="2 101", next_hop=self.AS2,
                           filtered=True, reject_reason=12)
        self.log_contains(self.rs, "prefix not in client's r_set - REJECTING " + self.DATA["AS101_no_rset"])

    def test_040_bad_prefixes_received_by_rs_no_asset(self):
        """{}: bad prefixes received by rs: origin not in AS-SET"""

        # AS102_no_asset is announced by an (hypothetical) AS102 to AS101,
        # and AS102 is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route(self.rs, self.DATA["AS102_no_asset"], self.AS1_1,
                           as_path="1 101 102", next_hop=self.AS1_1,
                           filtered=True, reject_reason=9)
        self.receive_route(self.rs, self.DATA["AS102_no_asset"], self.AS2,
                           as_path="2 101 102", next_hop=self.AS2,
                           filtered=True, reject_reason=9)
        self.log_contains(self.rs, "origin ASN [102] not in allowed as-sets - REJECTING " + self.DATA["AS102_no_asset"])

    def test_040_bad_prefix_not_ipv6_global_unicat(self):
        """{}: bad prefixes received by rs: not IPv6 global unicast space"""
        if "AS101_no_ipv6_gl_uni" not in self.DATA:
            # it's an IPv4-only scenario
            return
        self.receive_route(self.rs, self.DATA["AS101_no_ipv6_gl_uni"],
                           filtered=True, reject_reason=10)
        self.log_contains(self.rs, "prefix is not in IPv6 Global Unicast space - REJECTING " + self.DATA["AS101_no_ipv6_gl_uni"])

    def test_040_default_rejected_by_rs(self):
        """{}: bad prefixes received by rs: default route"""
        self.receive_route(self.rs, self.DATA["Default_route"],
                           other_inst=self.AS3,
                           filtered=True, reject_reason=(2, 10))
        if IPNetwork(self.DATA["Default_route"]).version == 4:
            msg = "prefix is bogon - REJECTING " + self.DATA["Default_route"]
        else:
            msg = "prefix is not in IPv6 Global Unicast space - REJECTING " + self.DATA["Default_route"]
        self.log_contains(self.rs, msg)

    def test_040_bad_prefixes_even_if_irrdb_whitelist(self):
        """{}: bad prefixes received by rs: IRRdb white-list"""

        self.receive_route(self.rs, self.DATA["AS1_whitel_2"], as_path="1 1000",
                           filtered=True, reject_reason=9)
        self.log_contains(self.rs, "origin ASN [1000] not in allowed as-sets - REJECTING " + self.DATA["AS1_whitel_2"])

        self.receive_route(self.rs, self.DATA["AS1_whitel_3"], as_path="1 1011",
                           filtered=True, reject_reason=12)
        self.log_contains(self.rs, "prefix not in client's r_set - REJECTING " + self.DATA["AS1_whitel_3"])

        self.receive_route(self.rs, self.DATA["AS1_whitel_6"], as_path="1 1011",
                           filtered=True, reject_reason=12)
        self.log_contains(self.rs, "prefix not in client's r_set - REJECTING " + self.DATA["AS1_whitel_6"])

    def test_040_bad_prefixes_rpki_roas_as_route_objects_failed(self):
        """{}: bad prefixes received by rs: RPKI ROAs as route objects failed"""

        # More specific than ROA
        prefix = self.DATA["AS101_roa_routeobj_2"]
        self.receive_route(self.rs, prefix, self.AS1_1, as_path="1 101",
                           filtered=True, reject_reason=12)
        self.receive_route(self.rs, prefix, self.AS2, as_path="2 101",
                           filtered=True, reject_reason=12)

        # ROA OK but origin ASN not authorized by AS-SET
        prefix = self.DATA["AS101_roa_routeobj_4"]
        self.receive_route(self.rs, prefix, self.AS1_1, as_path="1 101 105",
                           filtered=True, reject_reason=9)
        self.receive_route(self.rs, prefix, self.AS2, as_path="2 101 105",
                           filtered=True, reject_reason=9)

    def test_041_bad_prefixes_not_received_by_clients(self):
        """{}: bad prefixes not received by clients"""
        for prefix in (self.DATA["bogon1"],
                       self.DATA["local1"],
                       self.DATA["pref_len1"],
                       self.DATA["peer_as1"],
                       self.DATA["invalid_asn1"],
                       self.DATA["aspath_len1"],
                       self.DATA["AS2_nonclient_nexthop2"]):
            for inst in (self.AS2, self.AS3, self.AS4):
                with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                    self.receive_route(inst, prefix)

        # Among the clients, only AS3 is expected to not see the
        # following prefixes because AS1 and AS2
        # receive them on their session with AS101
        for prefix in (self.DATA["AS101_no_rset"],
                       self.DATA["AS102_no_asset"],
                       self.DATA["AS101_roa_routeobj_2"],
                       self.DATA["AS101_roa_routeobj_4"]):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(self.AS3, prefix)

    def test_045_rpki_valid_prefix(self):
        """{}: RPKI, valid prefix received by rs"""

        self.receive_route(self.rs, self.DATA["AS101_roa_valid1"], self.AS1_1,
                           as_path="1 101")
        self.receive_route(self.rs, self.DATA["AS101_roa_valid1"], self.AS2,
                           as_path="2 101")

    def test_045_rpki_invalid_prefix_asn(self):
        """{}: RPKI, invalid prefix (bad ASN) received by rs"""

        if isinstance(self.rs, BIRDInstance):
            ext_comm_rpki_invalid = ["generic:0x43000000:0x2"]
        else:
            ext_comm_rpki_invalid = []

        self.receive_route(self.rs, self.DATA["AS101_roa_invalid1"], self.AS1_1,
                           as_path="1 101", filtered=True,
                           ext_comms=ext_comm_rpki_invalid)
        self.receive_route(self.rs, self.DATA["AS101_roa_invalid1"], self.AS2,
                           as_path="2 101", filtered=True,
                           ext_comms=ext_comm_rpki_invalid)

    def test_045_rpki_invalid_prefix_length(self):
        """{}: RPKI, invalid prefix (bad length) received by rs"""

        if isinstance(self.rs, BIRDInstance):
            ext_comm_rpki_invalid = ["generic:0x43000000:0x2"]
        else:
            ext_comm_rpki_invalid = []

        self.receive_route(self.rs, self.DATA["AS101_roa_badlen"], self.AS1_1,
                           as_path="1 101", filtered=True,
                           ext_comms=ext_comm_rpki_invalid)
        self.receive_route(self.rs, self.DATA["AS101_roa_badlen"], self.AS2,
                           as_path="2 101", filtered=True,
                           ext_comms=ext_comm_rpki_invalid)

    def test_045_rpki_valid_prefix_propagated_to_clients(self):
        """{}: RPKI, valid prefix propagated to clients"""

        self.receive_route(self.AS3, self.DATA["AS101_roa_valid1"], self.rs)

    def test_045_rpki_invalid_prefixes_not_propagated_to_clients(self):
        """{}: RPKI, invalid prefix (bad ASN) not propagated to clients"""

        for pref_id in ("AS101_roa_invalid1", "AS101_roa_badlen"):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(self.AS3, self.DATA[pref_id])

    def test_045_blackhole_with_roa(self):
        """{}: RPKI, blackhole request for a covered prefix"""

        self.receive_route(self.rs, self.DATA["AS101_roa_blackhole"], self.AS1_1, as_path="1 101",
                           std_comms=["65535:666"], lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_roa_blackhole"], self.AS2, as_path="2 101",
                           std_comms=["65535:666"], lrg_comms=[])
        for inst in (self.AS1_1, self.AS2):
            self.log_contains(self.rs, "blackhole filtering request from {{inst}} - ACCEPTING {}".format(
                self.DATA["AS101_roa_blackhole"]), {"inst": inst})

        next_hop = self.DATA["blackhole_IP"]

        self.receive_route(self.AS3, self.DATA["AS101_roa_blackhole"], self.rs,
                           next_hop=next_hop,
                           std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_050_prefixes_from_AS101_received_by_its_upstreams(self):
        """{}: prefixes from AS101 received by its upstreams"""
        self.receive_route(self.AS1_1, self.DATA["AS101_good1"], self.AS101)
        self.receive_route(self.AS1_2, self.DATA["AS101_good1"], self.AS101)
        self.receive_route(self.AS2, self.DATA["AS101_good1"], self.AS101)

    def test_051_prefixes_from_AS101_received_by_rs(self):
        """{}: prefixes from AS101 received by rs"""

        # rs should receive these prefixes that AS101 announces to AS1 and AS2.
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_1, as_path="1 101", next_hop=self.AS1_1)
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS2, as_path="2 101", next_hop=self.AS2)

        # AS101 peers with AS1_2 on the same network of rs
        # and AS1_2 has not 'next-hop-self' on the session with rs:
        # next-hop received by rs is the one of AS101, so the prefix
        # should be filtered by the rs.
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_2,
                           as_path="1 101", next_hop=self.AS101,
                           filtered=True, reject_reason=5)

    def test_060_communities_as_seen_by_AS101_upstreams(self):
        """{}: bad communities as seen by AS101 upstreams"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS101_bad_std_comm"], self.AS101, std_comms=["65530:0"])
            self.receive_route(inst, self.DATA["AS101_bad_lrg_comm"], self.AS101, lrg_comms=["999:65530:0"])
            self.receive_route(inst, self.DATA["AS101_other_s_comm"], self.AS101, std_comms=["888:0"])
            self.receive_route(inst, self.DATA["AS101_other_l_comm"], self.AS101, lrg_comms=["888:0:0"])
            self.receive_route(inst, self.DATA["AS101_bad_good_comms"], self.AS101, std_comms=["777:0", "65530:1"], lrg_comms=["777:0:0", "999:65530:1"])

    def test_061_bad_communities_scrubbed_by_rs_std(self):
        """{}: bad communities scrubbed by rs (std)"""

        self.receive_route(self.rs, self.DATA["AS101_bad_std_comm"], self.AS1_1, std_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_bad_std_comm"], self.AS2, std_comms=[])

    def test_061_bad_communities_scrubbed_by_rs_lrg(self):
        """{}: bad communities scrubbed by rs (lrg)"""
        self.receive_route(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS1_1, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS2, lrg_comms=[])

    def test_062_other_communities_not_scrubbed_by_rs_std(self):
        """{}: other communities not scrubbed by rs (std)"""

        for inst in (self.AS1_1, self.AS2):
            self.receive_route(self.rs, self.DATA["AS101_other_s_comm"], inst, std_comms=["888:0"])
            self.receive_route(self.rs, self.DATA["AS101_bad_good_comms"], inst, std_comms=["777:0"])

    def test_062_other_communities_not_scrubbed_by_rs_lrg(self):
        """{}: other communities not scrubbed by rs (lrg)"""
        for inst in (self.AS1_1, self.AS2):
            self.receive_route(self.rs, self.DATA["AS101_other_l_comm"], inst, lrg_comms=["888:0:0"])
            self.receive_route(self.rs, self.DATA["AS101_bad_good_comms"], inst, lrg_comms=["777:0:0"])

    def test_070_blackhole_filtering_as_seen_by_rs_BLACKHOLE(self):
        """{}: blackhole filtering requests as seen by rs (BLACKHOLE)"""

        self.receive_route(self.rs, self.DATA["AS2_blackhole1"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=["65535:666"], lrg_comms=[])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole1"], {"AS2_1": self.AS2})

    def test_070_blackhole_filtering_as_seen_by_rs_std_cust(self):
        """{}: blackhole filtering requests as seen by rs (std cust)"""

        self.receive_route(self.rs, self.DATA["AS2_blackhole2"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=["65534:0"], lrg_comms=[], ext_comms=[])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole2"], {"AS2_1": self.AS2})

    def test_070_blackhole_filtering_as_seen_by_rs_lrg_cust(self):
        """{}: blackhole filtering requests as seen by rs (lrg cust)"""
        self.receive_route(self.rs, self.DATA["AS2_blackhole3"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=[], lrg_comms=["65534:0:0"])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole3"], {"AS2_1": self.AS2})

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_BLACKHOLE(self):
        """{}: blackholed prefixes as seen by enabled clients (BLACKHOLE)"""
        for inst in (self.AS1_1, self.AS3, self.AS4):
            self.receive_route(inst, self.DATA["AS2_blackhole1"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_std_cust(self):
        """{}: blackholed prefixes as seen by enabled clients (std_cust)"""
        for inst in (self.AS1_1, self.AS3, self.AS4):
            self.receive_route(inst, self.DATA["AS2_blackhole2"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_lrg_cust(self):
        """{}: blackholed prefixes as seen by enabled clients (lrg_cust)"""
        for inst in (self.AS1_1, self.AS3, self.AS4):
            self.receive_route(inst, self.DATA["AS2_blackhole3"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_not_seen_by_not_enabled_clients(self):
        """{}: blackholed prefixes not seen by not enabled clients"""

        # AS1_2 not enabled to receive blackholed prefixes
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole1"])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole2"])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole3"])
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole1"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole2"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole3"]), {"AS1_2": self.AS1_2})

    def test_075_gshut_enabled_client(self):
        """{}: gshut by an enabled client"""

        prefix = self.DATA["AS103_gshut_1"]

        # AS103's prefix as seen by its upstreams, AS1 and AS2
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, prefix, as_path="101 103")
        self.receive_route(self.AS2, prefix, as_path="101 101 103")

        # prefix as seen by the route server (AS1 performed gshut)
        self.receive_route(self.rs, prefix, as_path="1 101 103",
                           other_inst=self.AS1_1,
                           local_pref=5, std_comms=["65535:0"])
        self.receive_route(self.rs, prefix, as_path="2 101 101 103",
                           other_inst=self.AS2,
                           local_pref=100, std_comms=[])

        # prefix as seen by other clients, via AS2 (longest path)
        self.receive_route(self.AS4, prefix, as_path="2 101 101 103",
                           next_hop=self.AS2, std_comms=[])

    def test_075_gshut_not_enabled_client(self):
        """{}: gshut by a not enabled client"""

        prefix = self.DATA["AS103_gshut_2"]

        # AS103's prefix as seen by its upstreams, AS1 and AS2
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, prefix, as_path="101 101 103")
        self.receive_route(self.AS2, prefix, as_path="101 103")

        # prefix as seen by the route server (AS2 tried gshut but not enabled)
        self.receive_route(self.rs, prefix, as_path="1 101 101 103",
                           other_inst=self.AS1_1,
                           local_pref=100, std_comms=[])
        self.receive_route(self.rs, prefix, as_path="2 101 103",
                           other_inst=self.AS2,
                           local_pref=100, std_comms=[])

        # prefix as seen by other clients, via AS2 (still best path)
        self.receive_route(self.AS4, prefix, as_path="2 101 103",
                           next_hop=self.AS2, std_comms=[])

    def test_080_control_communities_AS1_only(self):
        """{}: control communities, announce to AS1 only"""

        self.receive_route(self.AS1_1, self.DATA["AS3_cc_AS1only"], self.rs,
                           as_path="3", next_hop=self.AS3,
                           std_comms=[], lrg_comms=[])
        self.receive_route(self.AS1_2, self.DATA["AS3_cc_AS1only"], self.rs,
                           as_path="3", next_hop=self.AS3,
                           std_comms=[], lrg_comms=[])

        for inst in (self.AS2, self.AS4):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS3_cc_AS1only"])
        self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{AS2}}".format(self.DATA["AS3_cc_AS1only"]), {"AS2": self.AS2})

    def test_080_control_communities_not_AS1(self):
        """{}: control communities, announce to all except AS1"""

        for inst in (self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_cc_not_AS1"], self.rs,
                               as_path="3", next_hop=self.AS3,
                               std_comms=[], lrg_comms=[])

        for inst in (self.AS1_1, self.AS1_2):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS3_cc_not_AS1"])
            self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_not_AS1"]), {"other_inst": inst})

    def test_080_control_communities_none(self):
        """{}: control communities, don't announce to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS3_cc_none"])
            self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_none"]), {"other_inst": inst})

    def test_081_control_communities_prepend1any(self):
        """{}: control communities, prepend once to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prepend1any"], self.rs, as_path="3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend2any(self):
        """{}: control communities, prepend twice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prepend2any"], self.rs, as_path="3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend3any(self):
        """{}: control communities, prepend thrice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prepend3any"], self.rs, as_path="3 3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend1_AS1(self):
        """{}: control communities, prepend once to AS1"""

        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, self.DATA["AS3_prepend1_AS1"], self.rs, as_path="3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])
        for inst in (self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prepend1_AS1"], self.rs, as_path="3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend2_AS2(self):
        """{}: control communities, prepend twice to AS2"""

        self.receive_route(self.AS2, self.DATA["AS3_prepend2_AS2"], self.rs, as_path="3 3 3",
                           next_hop=self.AS3, std_comms=[], lrg_comms=[])
        for inst in (self.AS1_1, self.AS1_2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prepend2_AS2"], self.rs, as_path="3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend3_AS1_1_others(self):
        """{}: control communities, prepend thrice to AS1, once to others"""

        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, self.DATA["AS3_prep3AS1_1any"], self.rs, as_path="3 3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])
        for inst in (self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_prep3AS1_1any"], self.rs, as_path="3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_083_control_communities_AS3_noexport_to_any(self):
        """{}: control communities, NO_EXPORT to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            self.receive_route(inst, self.DATA["AS3_noexport_any"], self.rs,
                               as_path="3", next_hop=self.AS3,
                               std_comms=["65535:65281"],
                               lrg_comms=[], ext_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS101, self.DATA["AS3_noexport_any"])

    def test_083_control_communities_AS3_noexport_to_AS1(self):
        """{}: control communities, NO_EXPORT to AS1"""

        pref = self.DATA["AS3_noexport_AS1"]
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, pref, self.rs, as_path="3", next_hop=self.AS3,
                               std_comms=["65535:65281"],
                               lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS2, pref, self.rs, as_path="3 3 3 3", next_hop=self.AS3,
                           std_comms=[],
                           lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS101, pref, as_path="2 3 3 3 3")
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS101, pref, as_path="1 3")

    def test_083_control_communities_AS3_rfc1997_noexport(self):
        """{}: control communities, RFC1997 NO_EXPORT"""

        pref = self.DATA["AS3_rfc1997_noexp"]
        for inst in (self.AS1_1, self.AS1_2, self.AS2, self.AS4):
            self.receive_route(inst, pref, self.rs, as_path="3",
                               next_hop=self.AS3, std_comms=["65535:65281"],
                               lrg_comms=[], ext_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS101, pref)

    def _test_084_AS1_1_and_AS1_2_only(self, pref):
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, pref, self.rs, as_path="4",
                               std_comms=[], lrg_comms=[], ext_comms=[])
        for inst in (self.AS2, self.AS3):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, pref)

    def test_084_control_communities_rtt_AS4_only_peers_lt_15(self):
        """{}: control communities, RTT, only peers <= 15 ms"""
        pref = self.DATA["AS4_rtt_1"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["0:999", "64532:15"])
        self._test_084_AS1_1_and_AS1_2_only(pref)

    def test_084_control_communities_rtt_AS4_only_peers_lt_5(self):
        """{}: control communities, RTT, only peers <= 5 ms"""
        pref = self.DATA["AS4_rtt_2"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["0:999", "64532:5"])
        self._test_084_AS1_1_and_AS1_2_only(pref)

    def test_084_control_communities_rtt_AS4_not_peers_gt_15(self):
        """{}: control communities, RTT, not peers > 15 ms"""
        pref = self.DATA["AS4_rtt_3"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["64531:15"])
        self._test_084_AS1_1_and_AS1_2_only(pref)

    def test_084_control_communities_rtt_AS4_not_peers_gt_5(self):
        """{}: control communities, RTT, not peers > 5 ms"""
        pref = self.DATA["AS4_rtt_4"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["64531:5"])
        self._test_084_AS1_1_and_AS1_2_only(pref)

    def test_084_control_communities_rtt_AS4_not_peers_gt_5_but_AS3(self):
        """{}: control communities, RTT, not peers > 5 ms + AS3"""
        pref = self.DATA["AS4_rtt_5"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["65501:3", "64531:5"])
        for inst in (self.AS1_1, self.AS1_2, self.AS3):
            self.receive_route(inst, pref, self.rs,
                               std_comms=[], lrg_comms=[], ext_comms=[])
        for inst in [self.AS2]:
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, pref)

    def test_084_control_communities_rtt_AS4_not_peers_lt_5_gt_100(self):
        """{}: control communities, RTT, not peers <= 5 and > 100 ms"""
        pref = self.DATA["AS4_rtt_6"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["64530:5", "64531:100"])
        for inst in [self.AS2]:
            self.receive_route(inst, pref, self.rs,
                               std_comms=[], lrg_comms=[], ext_comms=[])
        for inst in [self.AS1_1, self.AS1_2, self.AS3]:
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, pref)

    def test_084_control_communities_rtt_AS4_blackhole_not_to_peers_gt_20(self):
        """{}: control communities, RTT, blackhole, not peers > 20 ms"""
        expected_bh_next_hop = self.DATA["blackhole_IP"]

        pref = self.DATA["AS4_rtt_7"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["65535:666", "64531:20"])
        for inst in [self.AS1_1, self.AS2]:
            self.receive_route(inst, pref, self.rs,
                               next_hop=expected_bh_next_hop,
                               std_comms=["65535:666", "65535:65281"],
                               lrg_comms=[], ext_comms=[])
        for inst in [self.AS1_2, self.AS3]:
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, pref)

    def test_085_control_communities_rtt_prep3x_gt_100_2x_gt_10(self):
        """{}: control communities, RTT, prepend 3x > 100 ms, 2x > 10 ms"""
        pref = self.DATA["AS4_rtt_8"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["64539:100", "64538:10"])
        for inst in [self.AS1_1, self.AS1_2]:
            self.receive_route(inst, pref, self.rs, as_path="4",
                               std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS2, pref, self.rs, as_path="4 4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS3, pref, self.rs, as_path="999 4 4 4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])

    def test_085_control_communities_rtt_prep3x_lt_5_2x_lt_20_1x_any(self):
        """{}: control communities, RTT, prepend 3x <= 5 ms, 2x <= 20 ms, 1x any"""
        pref = self.DATA["AS4_rtt_9"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=["64536:5", "64535:20", "65521:65521"])
        for inst in [self.AS1_1, self.AS1_2]:
            self.receive_route(inst, pref, self.rs, as_path="4 4 4 4",
                               std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS2, pref, self.rs, as_path="4 4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS3, pref, self.rs, as_path="999 4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])

    def test_085_control_communities_rtt_ext_comms_prep1x_gt_10_2x_gt_20(self):
        """{}: control communities, RTT, ext comms, prepend 1x > 10 ms, 2x > 20 ms"""
        if isinstance(self.rs, BIRDInstance):
            ext_comm_rpki_unknown = ["generic:0x43000000:0x1"]
        else:
            ext_comm_rpki_unknown = []
        pref = self.DATA["AS4_rtt_10"]
        self.receive_route(self.rs, pref, self.AS4,
                           std_comms=[],
                           ext_comms=ext_comm_rpki_unknown + ["rt:64537:10",
                                                              "rt:64538:20"])
        for inst in [self.AS1_1, self.AS1_2]:
            self.receive_route(inst, pref, self.rs, as_path="4",
                               std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS2, pref, self.rs, as_path="4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS3, pref, self.rs, as_path="999 4 4 4",
                            std_comms=[], lrg_comms=[], ext_comms=[])

    def test_100_prefixes_received_by_clients_AS1_1(self):
        """{}: prefixes received by clients: AS1_1"""
        self.receive_route(self.AS1_1, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_1, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_1, self.DATA["AS2_nonclient_nexthop1"],
                           self.rs, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_100_prefixes_received_by_clients_AS1_2(self):
        """{}: prefixes received by clients: AS1_2"""

        self.receive_route(self.AS1_2, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_2, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_2, self.DATA["AS2_nonclient_nexthop1"],
                           self.rs, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_100_prefixes_received_by_clients_AS2(self):
        """{}: prefixes received by clients: AS2"""

        self.receive_route(self.AS2, self.DATA["AS1_good1"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.AS2, self.DATA["AS1_good2"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.AS2, self.DATA["AS1_good3"], self.rs, as_path="1", next_hop=self.AS1_1)

    def test_100_prefixes_received_by_clients_AS3(self):
        """{}: prefixes received by clients: AS3"""

        # AS3 has prepend_rs_as, so all the prefixes received from the rs
        # have AS_PATH "999 x"

        # prefix announced by AS1_2 only but with AS1_1 next-hop
        self.receive_route(self.AS3, self.DATA["AS1_good3"], self.rs, as_path="999 1", next_hop=self.AS1_1)

        # AS3 must receive prefixes from AS2
        self.receive_route(self.AS3, self.DATA["AS2_good1"], self.rs, as_path="999 2", next_hop=self.AS2)
        self.receive_route(self.AS3, self.DATA["AS2_good2"], self.rs, as_path="999 2", next_hop=self.AS2)

        # AS3 must receive prefixes from AS1 (this test only verifies
        # that the following prefixes reach AS3, without taking into account
        # AS_PATH nor NEXT_HOP; the "_with_ADD_PATH" version of this test
        # also verifies that these prefixes are received twice by AS3)
        for pref in ("AS101_good1", "AS101_bad_std_comm", "AS101_bad_lrg_comm",
                     "AS101_other_s_comm", "AS101_other_l_comm",
                     "AS101_bad_good_comms", "AS104_arin_1", "AS104_nicbr_1"):
            self.receive_route(self.AS3, self.DATA[pref], self.rs)

        for pref in ("AS101_roa_routeobj_1", "AS101_roa_routeobj_3"):
            self.receive_route(self.AS3, self.DATA[pref], self.rs)

    def test_100_prefixes_received_by_clients_AS3_with_ADD_PATH(self):
        """{}: prefixes received by clients: AS3 (with ADD-PATH)"""

        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("ADD-PATH not supported by OpenBGPD")

        # AS3 has prepend_rs_as, so all the prefixes received from the rs
        # have AS_PATH "999 x"

        # AS3 must receive prefixes from both AS1_1 and AS1_2 because the
        # session on the rs is configured with ADD-PATH
        self.receive_route(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_2)
        self.receive_route(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_2)

        # AS101 announces its prefixes to both AS1 (AS1_1 and AS1_2 clients) and AS2.
        # AS101 prefixes received by AS1_2 are rejected by rs because of next-hop-policy failure.
        # The other prefixes should be received twice by AS3 because ADD-PATH
        # is configured for the session on the rs.
        self.receive_route(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, std_comms=["888:0"])
        self.receive_route(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, std_comms=["888:0"])

        self.receive_route(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, lrg_comms=["888:0:0"])
        self.receive_route(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, lrg_comms=["888:0:0"])

        self.receive_route(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 1 101", next_hop=self.AS1_1,
                           std_comms=["777:0"], lrg_comms=["777:0:0"])
        self.receive_route(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 2 101", next_hop=self.AS2,
                           std_comms=["777:0"], lrg_comms=["777:0:0"])

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()
        self.test_020_sessions_up()

class BasicScenario_TagRejectPolicy(LiveScenario_TagRejectPolicy):

    def test_042_bad_prefixes_received_by_rs_bogon_wrong_tag(self):
        """{}: bad prefixes received by rs: bogon (wrong tag)"""
        with six.assertRaisesRegex(self, AssertionError, "real reasons 2, expected reason 1."):
            self.receive_route(self.rs, self.DATA["bogon1"], self.AS1_1,
                               as_path="1", next_hop=self.AS1_1,
                               filtered=True, reject_reason=1)

    def test_042_bad_prefixes_received_by_rs_global_blacklist_wrong_tag(self):
        """{}: bad prefixes received by rs: global blacklist (wrong tag)"""
        with six.assertRaisesRegex(self, AssertionError, "real reasons 3, expected reason 1."):
            self.receive_route(self.rs, self.DATA["local1"], self.AS1_1,
                               as_path="1", next_hop=self.AS1_1,
                               filtered=True, reject_reason=1)

class BasicScenarioBIRD(BasicScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    TARGET_VERSION = None
    IP_VER = None

    @classmethod
    def _get_local_file_name(cls):
        return "bird_local_file"

    @classmethod
    def _get_local_file(cls):
        local_filename = cls._get_local_file_name()

        res = []
        for ip_ver in (4, 6):
            if ip_ver == cls.IP_VER or cls.IP_VER is None:
                res.append(
                    (
                        cls.use_static_file("{}.local{}".format(local_filename, ip_ver)),
                        "/etc/bird/footer{}.local".format(ip_ver)
                    )
                )
        return res

    @classmethod
    def _setup_rs_instance(cls):
        if cls.IP_VER is None:
            ip_vers = [4, 6]
        else:
            ip_vers = [cls.IP_VER]

        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     target_version=cls.TARGET_VERSION,
                                     local_files=["footer{}".format(ip_ver) for ip_ver in ip_vers]),
                    "/etc/bird/bird.conf"
                )
            ] + cls._get_local_file(),
        )

class BasicScenarioBIRD2(BasicScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

    @classmethod
    def _get_local_file_name(cls):
        return "bird2_local_file"

class BasicScenarioOpenBGPD(BasicScenario_TagRejectPolicy, BasicScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder
    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     local_files_dir="/etc/bgpd",
                                     local_files=["post-clients", "post-filters"],
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                ),
                (
                    cls.use_static_file("openbgpd_post-clients.local"),
                    "/etc/bgpd/post-clients.local"
                ),
                (
                    cls.use_static_file("openbgpd_post-filters.local"),
                    "/etc/bgpd/post-filters.local"
                )
            ]
        )

class BasicScenarioOpenBGPDPrevious(BasicScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.BGP_SPEAKER_VERSION

class BasicScenarioOpenBGPDLatest(BasicScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.BGP_SPEAKER_VERSION



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import yaml

from .cfg_base import TestConfigParserBase 
from pierky.arouteserver.config.asns import ConfigParserASNS


class TestConfigParserASNs(TestConfigParserBase):

    FILE_PATH = "config.d/clients.yml"
    CONFIG_PARSER_CLASS = ConfigParserASNS
    SHORT_DESCR = "ASNs config parser"

    def test_valid_cfg(self):
        """{}: valid configuration"""
        self._contains_err()
        self.assertEqual(self.cfg["AS10745"]["as_sets"], None)
        self.assertEqual(self.cfg["AS3333"]["as_sets"], None)

    def test_invalid_as(self):
        """{}: invalid AS"""
        cfg = "\n".join([
            "asns:",
            "  as123:"
        ])
        self.load_config(yaml=cfg)
        self._contains_err("Invalid ASN format in 'asns' section for 'as123':")

        cfg = "\n".join([
            "asns:",
            "  AS_123:"
        ])
        self.load_config(yaml=cfg)
        self._contains_err("Invalid ASN format in 'asns' section for 'AS_123':")

        cfg = "\n".join([
            "asns:",
            "  AS:"
        ])
        self.load_config(yaml=cfg)
        self._contains_err("Invalid ASN format in 'asns' section for 'AS':")

    def test_valid_asset(self):
        """{}: valid AS-SET"""
        cfg = "\n".join([
            "asns:",
            "  AS123:",
            "    as_sets:",
            "      - 'AS-AS123'"  
        ])
        self.load_config(yaml=cfg)
        self._contains_err()

        cfg = "\n".join([
            "asns:",
            "  AS123:",
            "    as_sets:",
            "      - 'AS-AS123'",
            "      - 'AS-AS456'"
        ])
        self.load_config(yaml=cfg)
        self._contains_err()
        self.assertEqual(self.cfg["AS123"]["as_sets"][0], "AS-AS123")

    def test_invalid_asset(self):
        """{}: invalid AS-SET"""
        cfg = "\n".join([
            "asns:",
            "  AS123:",
            "    as_sets:",
            "      a: 'AS-AS123'"  
        ])
        self.load_config(yaml=cfg)
        self._contains_err("Error parsing 'as_sets' at 'asns' level - Invalid format: must be a list.")



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from pierky.arouteserver.tests.mocked_env import MockedEnv
from pierky.arouteserver.peering_db import PeeringDBNet
from pierky.arouteserver.errors import PeeringDBNoInfoError 


class TestPeeringDBInfo(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        MockedEnv(default=False, peering_db=True,
                  base_dir=os.path.dirname(__file__))

    @classmethod
    def tearDownClass(cls):
        MockedEnv.stopall()

    def test_net1(self):
        """PeeringDB network: get data"""
        net = PeeringDBNet(1)
        net.load_data()
        self.assertEqual(net.info_prefixes4, 20)
        self.assertEqual(net.info_prefixes6, 10)

    def test_no_data(self):
        """PeeringDB network: missing data"""
        net = PeeringDBNet(2)
        with self.assertRaises(PeeringDBNoInfoError):
            net.load_data()

    def test_parse_as_sets(self):
        """PeeringDB: AS-SETs parsing"""
        net = PeeringDBNet(1)
        net.load_data()
        self.assertEqual(net.parse_as_sets("AS-1"), ["AS-1"])
        self.assertEqual(net.parse_as_sets("AS-1, AS-2, AA:BB"), ["AS-1", "AS-2"])
        self.assertEqual(net.parse_as_sets("AS-1 / AS-2\nAS-3"), ["AS-1", "AS-2", "AS-3"])
        self.assertEqual(net.parse_as_sets("AS-"), [])
        self.assertEqual(net.parse_as_sets("AS-A_BB, as-b"), ["AS-A_BB", "AS-B"])
        self.assertEqual(net.parse_as_sets("AA-B"), [])
        self.assertEqual(net.parse_as_sets("AA:AS-BB"), [])
        self.assertEqual(net.parse_as_sets("AS:AS-BB"), [])
        self.assertEqual(net.parse_as_sets("AS1:AS-BB"), ["AS1:AS-BB"])
        self.assertEqual(net.parse_as_sets("AS1"), ["AS1"])
        self.assertEqual(net.parse_as_sets("RIPE:AS-B"), ["RIPE::AS-B"])
        self.assertEqual(net.parse_as_sets("UNKNOWNSOURCE:AS-B"), [])
        self.assertEqual(net.parse_as_sets("UNKNOWNSOURCE::AS-B"), ["UNKNOWNSOURCE::AS-B"])
        self.assertEqual(net.parse_as_sets("RIPE::AS-B"), ["RIPE::AS-B"])
        self.assertEqual(net.parse_as_sets("RIPE: AS-B"), ["AS-B"])
        self.assertEqual(net.parse_as_sets("RIPE: AS1:AS-B:AS2, ARIN: AS1:AS-1-CUST"), ["AS1:AS-B:AS2", "AS1:AS-1-CUST"])
        self.assertEqual(net.parse_as_sets("ipv4:AS-B"), ["AS-B"])
        self.assertEqual(net.parse_as_sets("ipv4:RIPE::AS-B"), ["RIPE::AS-B"])
        self.assertEqual(net.parse_as_sets("ripe: AS-A, RaDB: AS-ONE"), ["AS-A", "AS-ONE"])
        self.assertEqual(net.parse_as_sets("AS-ONE@RIPE"), ["RIPE::AS-ONE"])
        self.assertEqual(net.parse_as_sets("RIPE::AS-ONE@RIPE"), [])
        self.assertEqual(net.parse_as_sets("RIPE:AS-ONE@RIPE"), [])
        self.assertEqual(net.parse_as_sets("RIPE: AS-ONE@RIPE"), ["RIPE::AS-ONE"])
        self.assertEqual(net.parse_as_sets("AS-ONE@TWO@RIPE"), [])



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from aggregate6 import aggregate
import logging
import os
from packaging import version
import sys
import time
import yaml

from jinja2 import Environment, FileSystemLoader, StrictUndefined

from .config.general import ConfigParserGeneral
from .config.bogons import ConfigParserBogons
from .config.asns import ConfigParserASNS
from .config.clients import ConfigParserClients
from .enrichers.arin_db_dump import ARINWhoisDBDumpEnricher
from .enrichers.registrobr_db_dump import RegistroBRWhoisDBDumpEnricher
from .enrichers.irrdb import IRRDBConfigEnricher_ASNs, \
                             IRRDBConfigEnricher_Prefixes
from .enrichers.pdb_as_set import PeeringDBConfigEnricher_ASSet
from .enrichers.pdb_max_prefix import PeeringDBConfigEnricher_MaxPrefix
from .enrichers.pdb_never_via_route_servers import NeverViaRouteServersEnricher
from .enrichers.rpki_roas import RPKIROAsEnricher
from .enrichers.rtt import RTTGetterConfigEnricher
from .errors import MissingDirError, MissingFileError, BuilderError, \
                    ARouteServerError, MissingArgumentError, \
                    TemplateRenderingError, CompatibilityIssuesError, \
                    ConfigError, MissingGeneralConfigFileError
from .ipaddresses import IPNetwork
from .irrdb import IRRDBInfo
from .cached_objects import CachedObject, normalize_expiry_time


class ConfigBuilder(object):
    """The base configuration builder class.

    This class must be derived by BGP-daemon-specific classes.
    """

    LOCAL_FILES_IDS = None
    LOCAL_FILES_BASE_DIR = None

    DEFAULT_VERSION = None

    IGNORABLE_ISSUES = []

    REJECT_REASONS = {
        "1": "Invalid AS_PATH length",
        "2": "Prefix is bogon",
        "3": "Prefix is in global blacklist",
        "4": "Invalid AFI",
        "5": "Invalid NEXT_HOP",
        "6": "Invalid left-most ASN",
        "7": "Invalid ASN in AS_PATH",
        "8": "Transit-free ASN in AS_PATH",
        "9": "Origin ASN not in IRRDB AS-SETs",
        "10": "IPv6 prefix not in global unicast space",
        "11": "Prefix is in client blacklist",
        "12": "Prefix not in IRRDB AS-SETs",
        "13": "Invalid prefix length",
        "14": "RPKI INVALID route",
        "15": "Never via route-servers ASN in AS_PATH",
    }

    def validate_bgpspeaker_specific_configuration(self):
        """Check compatibility between config and target BGP speaker

        Returns:
            True if there are no compatibility issues;
            False if there are compatibility issues that can be acknowledged
            via command line argument.
        Raises exception in case of blocking errors.
        """
        return True

    def process_bgpspeaker_specific_compatibility_issue(self, issue_id, text):
        """Process a compatibility issue with the target BGP speaker

        If the issue has been ignored via command-line, logs a warning
        message and returns True.
        Otherwise, logs an error message and returns False.

        """
        assert issue_id in self.IGNORABLE_ISSUES

        msg = "Compatibility issue ID '{}'. {}".format(issue_id, text)
        if issue_id in self.ignore_errors or "*" in self.ignore_errors:
            logging.warning(
                "{} - Ignored".format(msg)
            )
            return True
        logging.error(msg)
        return False

    def enrich_j2_environment(self, env):
        pass

    @staticmethod
    def _get_cfg(obj_or_path, cls, descr, **kwargs):
        assert obj_or_path is not None
        if isinstance(obj_or_path, cls):
            return obj_or_path
        elif isinstance(obj_or_path, str):
            path = os.path.expanduser(obj_or_path)
            if not os.path.isfile(path):
                raise MissingFileError(path)
            obj = cls(**kwargs)
            try:
                obj.load(path)
            except ARouteServerError as e:
                raise BuilderError(
                    "One or more errors occurred while loading "
                    "{} configuration file{}".format(
                        descr, ": {}".format(str(e)) if str(e) else ""
                    )
                )
            return obj

    @staticmethod
    def _check_is_dir(arg, path):
        if path is None:
            raise MissingArgumentError(arg)
        dir_path = os.path.expanduser(path)
        if not os.path.isdir(dir_path):
            raise MissingDirError(path)
        return dir_path

    def __init__(self, template_dir=None, template_name=None,
                 cache_dir=None, cache_expiry=CachedObject.DEFAULT_EXPIRY,
                 bgpq3_path="bgpq3", bgpq3_host=IRRDBInfo.BGPQ3_DEFAULT_HOST,
                 bgpq3_sources=IRRDBInfo.BGPQ3_DEFAULT_SOURCES,
                 rtt_getter_path=None, threads=4,
                 ip_ver=None, perform_graceful_shutdown=False,
                 ignore_errors=[], live_tests=False,
                 local_files=[], local_files_dir=None, target_version=None,
                 cfg_general=None, cfg_bogons=None, cfg_clients=None,
                 **kwargs):
        """Initialize the configuration builder.

        Here, external data sources are also queried to enrich the
        configuration with additional data (PeeringDB records, ASNs and
        prefixes from IRRDBs, ...).

        Raises:

            ARouteServerError or derived exceptions
              (from pierky.arouteserver.errors).

              Exceptions raised here can have no arguments and their string
              representation can be empty: in these cases, it means that
              one or more errors have been logged using the ``logging`` module.

        Args:

            template_dir (str): the directory that contains the templates
                that must be used to render the output configuration.

                Example: /home/user/arouteserver/templates/bird

                Same of:

                - *--templates-dir* CLI argument.
                - *templates_dir* program's configuration file option.

            template_name (str): the name of the file that must be used to
                render the output configuration.

                Example: main.j2

                Same of:

                - *--template-file-name* CLI argument.
                - *template_name* program's configuration file option.

            cfg_general (str)
            cfg_clients (str)
            cfg_bogons (str): paths to the YAML
                files containing the general route server policy, the clients
                list and the list of bogon prefixes.

                Example: ``cfg_general="/home/user/arouteserver/general.yml"``

                Same of:

                - *--general*, *--clients*, *--bogons* CLI arguments.
                - *cfg_general*, *cfg_clients*, *cfg_bogons* program's
                  configuration file options.

            cache_dir (str): the directory that will be used to store results
                from external data sources queries (PeeringDB info, IRRDBs).

                Same of:

                - *--cache-dir* CLI argument.
                - *cache_dir* program's configuration file option.

            cache_expiry (int or dict): how long cached data must be considered
                valid, in seconds. Each "cacheable object" (PeeringDB info,
                IRR datasets, ...) can have its own expiry time. If an int is
                given here, all the expiry times will have the same duration,
                otherwise cacheable objects will pick their specific value
                or use the 'general' one if no more specific value is given.

                Same of:

                - *cache_expiry* program's configuration file option.

            ip_ver (int): if *None*, the output configuration will be targeted
                for both IPv4 and IPv6; otherwise, set this to *4* or to
                *6* to obtain AFI-specific output configuration.

                Same of:

                - *--ip-ver* CLI argument.

            perform_graceful_shutdown (bool): when True, the output config
                includes an outbound policy which is applied to BGP
                sessions toward the clients and which adds the
                GRACEFUL_SHUTDOWN BGP community (65535:0) to all the
                routes that the route server announces to them.

                Same of:

                - *--perform-graceful-shutdown* CLI argument.

            target_version (str): the BGP daemon target version for which the
                output configuration must be generated.

                This is used to detect and/or solve any compatibility issue
                with some features that are available only using a specific
                version of the target BGP daemon.

                The list of available versions is taken from the derived BGP
                daemon specific classes' ``AVAILABLE_VERSION`` attribute.

                The default value is taken from the derived BGP daemon
                specific classes' ``DEFAULT_VERSION`` attribute.

                Example: on OpenBGPD, to avoid errors when building configs
                that use large BGP communities (available only on
                OpenBGPD/OpenBSD > 6.1) use ``target_version="6.1"``

                Same of:

                - *--target-version* CLI argument.

            ignore_errors (list): a list of issue IDs (strings) that must be
                ignored when building the configuration.

                Depending on the target BGP daemon and its version, some
                features may be unavailable; ARouteServer produces errors
                when one or more of these features are enabled in the route
                server configuration YAML file. These errors are marked with
                an 'issue ID' that can be reported in this list to instruct
                ARouteServer to ignore it and to continue the building process.

                Use ``ignore_errors=["*"]`` to ignore any issue.

                Example: ``ignore_errors=["add_path"]`` to ignore the issue due
                to the lack of support for ADD_PATH in OpenBGPD.

                Same of:

                - *--ignore-issues* CLI argument.

            local_files (list): the list of local files IDs for which the
                relative inclusion point must be enabled on the output
                configuration. Details: https://arouteserver.readthedocs.io/en/latest/CONFIG.html#site-specific-custom-configuration-files

                The list of available IDs is taken from the derived BGP daemon
                specific classes' ``LOCAL_FILES_IDS`` attribute.

                Example: ``local_files=["header4", "footer4"]``

                Same of:

                - *--use-local-files* CLI argument.

            local_files_dir (str): the base directory of the local files that
                will be included in the output configuration.

                The default value is taken from the derived BGP daemon
                specific classes' ``LOCAL_FILES_BASE_DIR`` attribute.

                Example: /etc/bird

                Same of:

                - *--local-files-dir* CLI argument.

            bgpq3_path (str): path to the 'bgpq3' external program; this will
                be used to expand AS macros and to obtain the list of
                authorized origin ASNs and prefixes from IRRDBs.

                Same of:

                - *bgpq3_path* program's configuration file option.

            bgpq3_host (str): the host that will be queried by bgpq3; this
                will be used to set the *-h* argument of the program.

                Same of:

                - *bgpq3_host* program's configuration file option.

            bgpq3_sources (str): a comma separated list of sources that will
                be used by the bgpq3 program; this will be used to set the
                *-S* argument of bgpq3.

                Same of:

                - *bgpq3_sources* program's configuration file option.

            rtt_getter_path (str): path to the program that is executed to
                determine the RTT of a peer.
                Syntax and details can be found at the following URL:
                https://arouteserver.readthedocs.io/en/latest/RTT_GETTER.html

                Same of:

                - *rtt_getter_path* program's configuration file option.

            threads (int): number of concurrent threads used to gather
                additional data from external sources (bgpq3, PeeringDB, ...)

                Same of:

                - *threads* program's configuration file option.

            kwargs: additional arguments used by BGP daemon specific builder
                classes.

            live_tests (bool): only used on live tests.
        """

        # Parameters initialization

        self.template_dir = self._check_is_dir(
            "template_dir", template_dir
        )

        self.template_name = template_name
        if not self.template_name:
            raise MissingArgumentError("template_name")

        self.template_path = os.path.join(self.template_dir,
                                          self.template_name)
        if not os.path.isfile(self.template_path):
            raise MissingFileError(self.template_path)

        self.cache_dir = self._check_is_dir(
            "cache_dir", cache_dir
        )

        self.cache_expiry = normalize_expiry_time(cache_expiry)

        self.bgpq3_path = bgpq3_path
        self.bgpq3_host = bgpq3_host
        self.bgpq3_sources = bgpq3_sources

        self.rtt_getter_path = rtt_getter_path

        self.threads = threads

        try:
            with open(os.path.join(self.cache_dir, "write_test"), "w") as f:
                f.write("OK")
        except Exception as e:
            raise BuilderError(
                "Can't write into cache dir {}: {}".format(
                    self.cache_dir, str(e)
                )
            )

        self.ip_ver = ip_ver
        if self.ip_ver is not None:
            self.ip_ver = int(self.ip_ver)
            if self.ip_ver not in (4, 6):
                raise BuilderError("Invalid IP version: {}".format(ip_ver))

        self.perform_graceful_shutdown = perform_graceful_shutdown

        self.ignore_errors = ignore_errors or []

        self.live_tests = live_tests

        self.local_files = local_files
        self.local_files_dir = local_files_dir

        self.target_version = target_version or self.DEFAULT_VERSION

        try:
            self.cfg_general = self._get_cfg(cfg_general,
                                             ConfigParserGeneral,
                                             "general")
        except MissingFileError as e:
            raise MissingGeneralConfigFileError(e.path)

        self.cfg_bogons = self._get_cfg(cfg_bogons,
                                        ConfigParserBogons,
                                        "bogons")

        if isinstance(cfg_clients, str):
            self.cfg_asns = self._get_cfg(cfg_clients,
                                            ConfigParserASNS,
                                            "asns")
        else:
            self.cfg_asns = ConfigParserASNS()
            self.cfg_asns._load_from_yaml("{}")
            self.cfg_asns.parse()

        self.cfg_clients = self._get_cfg(cfg_clients,
                                         ConfigParserClients,
                                         "clients",
                                         general_cfg=self.cfg_general)

        self.kwargs = kwargs

        # Initially None; is set to IRRDB() and finally populated by
        # the IRRDB enrichers.
        # { "<as_set_bundle_id>": <IRRDBRecord>, ... }
        self.irrdb_info = None

        # { "<len>": [{"prefix": "<ip>/<len>", "max_len": x, "asn": "AS<n>"}]
        self.rpki_roas = {}

        # [<asn (int)>]
        self.never_via_route_servers_asns = []

        # { "<origin_asn>": ["a/b", "c/d"] }
        self.arin_whois_records = {}
        self.registrobr_whois_records = {}

        # Validation

        if self.local_files:
            if not isinstance(self.local_files, list):
                raise BuilderError(
                    "local_files must be a list of .local files IDs"
                )
            for local_file_id in self.local_files:
                if local_file_id not in self.LOCAL_FILES_IDS:
                    raise BuilderError(
                        "The .local file ID '{}' is invalid.".format(
                            local_file_id
                        )
                    )

        if not self.validate_bgpspeaker_specific_configuration():
            raise CompatibilityIssuesError(
                "One or more compatibility issues have been found."
            )

        # Processing

        logging.info("Started processing configuration "
                     "for {}".format(self.template_path))

        start_time = int(time.time())

        self.enrich_config()

        stop_time = int(time.time())

        logging.info("Configuration processing completed after "
                     "{} seconds.".format(stop_time - start_time))

    def enrich_config(self):
        errors = False

        # Unique ASNs from clients list.
        clients_asns = {}

        for client in self.cfg_clients.cfg["clients"]:
            # Unique ASNs
            asn = "AS{}".format(client["asn"])
            if asn in clients_asns:
                clients_asns[asn] += 1
            else:
                clients_asns[asn] = 1

            # Client's ID
            # Set 'id' as ASx_y where
            # x = ASN
            # y = progressive counter of clients per ASN
            client_id = "{}_{}".format(
                asn, clients_asns[asn]
            )
            client["id"] = client_id

        # BGP communities metadata
        for comm_name in ConfigParserGeneral.COMMUNITIES_SCHEMA:
            comm = ConfigParserGeneral.COMMUNITIES_SCHEMA[comm_name]
            self.cfg_general["communities"][comm_name]["type"] = comm["type"]
            self.cfg_general["communities"][comm_name]["peer_as"] = comm.get("peer_as", False)

        # Enrichers
        # Order matters: AS-SET from PeeringDB must be run first
        # in order to acquire missing AS-SETs that are processed
        # later by IRRDB enrichers. RPKI ROAs (when only used as
        # route objects) are processed only
        # for those origin ASNs that have been gathered from AS-SETs.
        filtering = self.cfg_general["filtering"]
        irrdb_cfg = filtering["irrdb"]
        used_enricher_classes = []

        if irrdb_cfg["peering_db"]:
            used_enricher_classes += [PeeringDBConfigEnricher_ASSet]

        used_enricher_classes += [IRRDBConfigEnricher_ASNs,
                                  IRRDBConfigEnricher_Prefixes,
                                  PeeringDBConfigEnricher_MaxPrefix]

        if self.cfg_general.rtt_based_functions_are_used:
            used_enricher_classes.append(RTTGetterConfigEnricher)

        if self.cfg_general.rpki_roas_needed and \
            self.cfg_general["rpki_roas"]["source"] == \
                "ripe-rpki-validator-cache":
            used_enricher_classes.append(RPKIROAsEnricher)

        if irrdb_cfg["use_arin_bulk_whois_data"]["enabled"]:
            used_enricher_classes.append(ARINWhoisDBDumpEnricher)

        if irrdb_cfg["use_registrobr_bulk_whois_data"]["enabled"]:
            used_enricher_classes.append(RegistroBRWhoisDBDumpEnricher)

        self.never_via_route_servers_asns = list(
            set(filtering["never_via_route_servers"]["asns"] or [])
        )
        if filtering["never_via_route_servers"]["peering_db"]:
            used_enricher_classes.append(NeverViaRouteServersEnricher)

        for enricher_class in used_enricher_classes:
            enricher = enricher_class(self, threads=self.threads)
            try:
                enricher.enrich()
            except ARouteServerError as e:
                errors = True
                if str(e):
                    logging.error(str(e))

        if errors:
            raise BuilderError()

    def _include_local_file(self, local_file_id):
        raise NotImplementedError()

    def render_template(self, output_file=None):
        """Render the output configuration.

        Raises:

            ARouteServerError or derived exceptions
              (from pierky.arouteserver.errors).

              Exceptions raised here can have no arguments and their string
              representation can be empty: in these cases, it means that
              one or more errors have been logged using the ``logging`` module.

        Args:

            output_file (file): the output file where the configuration must
                be written.
        """

        def sorted_rpki_roas():
            """Returns a list of ROAs, sorted by prefix length, prefix, ASN"""
            res = []
            prefix_lengths = sorted(map(int, self.rpki_roas.keys()))
            for pref_len in prefix_lengths:
                for roa in sorted(self.rpki_roas[str(pref_len)],
                                  key=lambda r: (r["prefix"], r["asn"])):
                    res.append(roa)
            return res

        self.data = {}
        self.data["ip_ver"] = self.ip_ver
        if self.ip_ver is None:
            self.data["list_ip_vers"] = [4, 6]
        else:
            self.data["list_ip_vers"] = [self.ip_ver]
        self.data["cfg"] = self.cfg_general
        self.data["bogons"] = self.cfg_bogons
        self.data["clients"] = self.cfg_clients
        self.data["asns"] = self.cfg_asns
        self.data["irrdb_info"] = self.irrdb_info
        self.data["rpki_roas"] = sorted_rpki_roas()
        self.data["arin_whois_records"] = self.arin_whois_records
        self.data["registrobr_whois_records"] = self.registrobr_whois_records
        self.data["never_via_route_servers_asns"] = self.never_via_route_servers_asns
        self.data["live_tests"] = self.live_tests
        self.data["rtt_based_functions_are_used"] = \
            self.cfg_general.rtt_based_functions_are_used
        self.data["perform_graceful_shutdown"] = self.perform_graceful_shutdown
        self.data["reject_reasons"] = self.REJECT_REASONS

        def ipaddr_ver(ip):
            return IPNetwork(ip).version

        def current_ipver(ip):
            if self.ip_ver is None:
                return True
            return IPNetwork(ip).version == self.ip_ver

        def is_ipver(data, ip_ver):
            prefix = data
            return IPNetwork(prefix).version == ip_ver

        def include_local_file(local_file_id):
            # The 'rpki_rtr_config' local_file_id is always allowed
            # to be included, because it's referenced directly in
            # the Jinja2 template for RPKI configuration.
            if local_file_id == "rpki_rtr_config":
                return self._include_local_file(local_file_id)

            if local_file_id not in self.LOCAL_FILES_IDS:
                raise AssertionError(
                    "Local file ID '{}' is referenced in J2 "
                    "templates but is not in LOCAL_FILES_IDS.".format(
                        local_file_id
                    )
                )
            local_files = self.local_files or []
            if local_file_id in local_files:
                return self._include_local_file(local_file_id)
            return ""

        def target_version_ge(v):
            if self.target_version:
                return version.parse(self.target_version) >= version.parse(v)
            return False

        def target_version_le(v):
            if self.target_version:
                return version.parse(self.target_version) <= version.parse(v)
            return False

        def get_normalized_rtt(v):
            if not v:
                return 0
            if v < 1:
                return 1
            if v > 60000:
                return 60000
            return int(round(v))

        def community_is_set(comm):
            if not comm:
                return False
            if not comm["std"] and not comm["lrg"] and not comm["ext"]:
                return False
            return True

        env = Environment(
            loader=FileSystemLoader(self.template_dir),
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=StrictUndefined
        )
        env.tests["current_ipver"] = current_ipver
        env.tests["is_ipver"] = is_ipver
        env.filters["community_is_set"] = community_is_set
        env.filters["ipaddr_ver"] = ipaddr_ver
        env.filters["include_local_file"] = include_local_file
        env.filters["target_version_ge"] = target_version_ge
        env.filters["target_version_le"] = target_version_le
        env.filters["get_normalized_rtt"] = get_normalized_rtt

        self.enrich_j2_environment(env)

        tpl = env.get_template(self.template_name)

        start_time = int(time.time())

        logging.info("Started template rendering "
                     "for {}".format(self.template_path))
        try:
            if output_file:
                for buf in tpl.generate(self.data):
                    output_file.write(buf)
            else:
                return tpl.render(self.data)
        except Exception as e:
            _, _, traceback = sys.exc_info()
            raise TemplateRenderingError(
                "Error while rendering template: {}".format(str(e)),
                traceback=traceback
            )
        finally:
            stop_time = int(time.time())

            logging.info("Template rendering completed after "
                        "{} seconds.".format(stop_time - start_time))

class BIRDConfigBuilder(ConfigBuilder):
    """BIRD configuration builder.

    The ``kwargs`` parameter of the ``__init__`` method can be used
    to pass the following additional arguments.

    Args:

        hooks (list): list of hook IDs for which to enable hooks in
            the output configuration. Details: https://arouteserver.readthedocs.io/en/latest/CONFIG.html#bird-hooks
    """

    LOCAL_FILES_IDS = ["header", "header4", "header6",
                       "footer", "footer4", "footer6",
                       "client", "client4", "client6"]
    LOCAL_FILES_BASE_DIR = "/etc/bird"

    HOOKS = ["pre_receive_from_client", "post_receive_from_client",
             "pre_announce_to_client", "post_announce_to_client",
             "route_can_be_announced_to", "announce_rpki_invalid_to_client",
             "scrub_communities_in", "scrub_communities_out",
             "apply_blackhole_filtering_policy"]

    AVAILABLE_VERSION = ["1.6.3", "1.6.4", "1.6.6", "1.6.7", "1.6.8",
                         "2.0.7"]
    DEFAULT_VERSION = "1.6.8"

    def validate_bgpspeaker_specific_configuration(self):
        if self.ip_ver is None and \
           version.parse(self.target_version) < version.parse("2.0"):
            raise BuilderError(
                "An explicit target IP version is needed "
                "to build BIRD 1.x configuration. Use the "
                "--ip-ver command line argument to supply one."
            )

        hooks = self.kwargs.get("hooks", [])
        if hooks:
            if not isinstance(hooks, list):
                raise BuilderError(
                    "hooks must be a list of hook names"
                )

        return True

    def _include_local_file(self, local_file_id):
        return 'include "{}";\n\n'.format(
            os.path.join(
                self.local_files_dir or self.LOCAL_FILES_BASE_DIR,
                "{}.local".format(local_file_id)
            )
        )

    def enrich_j2_environment(self, env):

        def hook_is_set(hook_name):
            if hook_name not in self.HOOKS:
                raise AssertionError(
                    "Hook '{}' is referenced in J2 "
                    "templates but is not in HOOKS.".format(hook_name)
                )
            hooks = self.kwargs.get("hooks", []) or []
            return hook_name in hooks

        env.filters["hook_is_set"] = hook_is_set

class OpenBGPDConfigBuilder(ConfigBuilder):
    """OpenBGPD configuration builder.
    """

    LOCAL_FILES_IDS = ["header",
                       "pre-irrdb", "post-irrdb",
                       "pre-clients", "post-clients", "client",
                       "pre-filters", "post-filters",
                       "footer"]
    LOCAL_FILES_BASE_DIR = "/etc/bgpd"

    AVAILABLE_VERSION = ["6.0", "6.1", "6.2", "6.3", "6.4", "6.5", "6.6"]
    DEFAULT_VERSION = "6.5"

    IGNORABLE_ISSUES = ["path_hiding", "transit_free_action",
                        "add_path", "max_prefix_action",
                        "blackhole_filtering_rewrite_ipv6_nh",
                        "large_communities", "extended_communities",
                        "graceful_shutdown", "internal_communities",
                        "rpki_roas_as_route_objects_source",
                        "rpki_roas_source"]

    def _include_local_file(self, local_file_id):
        return 'include "{}"\n\n'.format(
            os.path.join(
                self.local_files_dir or self.LOCAL_FILES_BASE_DIR,
                "{}.local".format(local_file_id)
            )
        )

    def validate_bgpspeaker_specific_configuration(self):
        res = True

        if self.cfg_general["path_hiding"]:
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "path_hiding",
                "The 'path_hiding' general configuration parameter is "
                "set to True, but the configuration generated by "
                "ARouteServer for OpenBGPD does not support "
                "path-hiding mitigation techniques."
            ):
                res = False

        transit_free_action = self.cfg_general["filtering"]["transit_free"]["action"]
        if transit_free_action and transit_free_action != "reject":
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "transit_free_action",
                "Transit free ASNs policy is configured with "
                "'action' = '{}' but only 'reject' is supported "
                "for OpenBGPD.".format(transit_free_action)
            ):
                res = False

        add_path_clients = []
        max_prefix_action_clients = []
        for client in self.cfg_clients.cfg["clients"]:
            if client["cfg"]["add_path"]:
                add_path_clients.append(client["ip"])

            max_prefix_action = client["cfg"]["filtering"]["max_prefix"]["action"]
            if max_prefix_action:
                if max_prefix_action not in ("shutdown", "restart"):
                    max_prefix_action_clients.append(client["ip"])

        if add_path_clients:
            clients = add_path_clients
            cnt = len(clients)
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "add_path",
                "ADD_PATH not supported by OpenBGPD but "
                "enabled for the following clients: {}{}.".format(
                    ", ".join(clients[:3]),
                    "" if cnt <= 3 else " and {} more".format(cnt - 3)
                )
            ):
                res = False

        if max_prefix_action_clients:
            clients = max_prefix_action_clients
            cnt = len(clients)
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "max_prefix_action",
                "Invalid max-prefix 'action' for the following "
                "clients: {}{}; only 'shutdown' and 'restart' "
                "are supported by OpenBGPD.".format(
                    ", ".join(clients[:3]),
                    "" if cnt <= 3 else " and {} more".format(cnt - 3)
                )
            ):
                res = False

        if self.cfg_general["blackhole_filtering"]["policy_ipv6"] == "rewrite-next-hop" and \
            version.parse(self.target_version) < version.parse("6.1"):
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "blackhole_filtering_rewrite_ipv6_nh",
                "On OpenBSD < 6.1 there is an issue related to next-hop rewriting "
                "that impacts blackhole filtering policies when "
                "'blackhole_filtering.policy_ipv6' is 'rewrite-next-hop': "
                "https://github.com/pierky/arouteserver/issues/3 "
                "If the release running on the route server includes the fix "
                "for this issue, please consider to enable this feature by "
                "setting the configuration target version to a value "
                "greater than or equal to '6.1' (--target-version command "
                "line argument)."
            ):
                res = False

        only_large_comms = []
        large_comms_used = []
        peer_as_ext_comms = []
        for comm_name in ConfigParserGeneral.COMMUNITIES_SCHEMA:
            comm = self.cfg_general["communities"][comm_name]

            # large comms used
            if comm["lrg"]:
                large_comms_used.append((comm_name, comm["lrg"]))

                # only large comms
                if not comm["std"] and not comm["ext"]:
                    only_large_comms.append((comm_name, comm["lrg"]))

            # peer_as ext communities not scrubbed
            comm_def = ConfigParserGeneral.COMMUNITIES_SCHEMA[comm_name]
            peer_as = comm_def.get("peer_as", False)
            direction = comm_def.get("type", None)

            if peer_as and direction == "inbound" and comm["ext"]:
                peer_as_ext_comms.append((comm_name, comm["ext"]))

        large_comms_advice = ("If large BGP communities are supported by the "
                              "release of OpenBGPD running on the route "
                              "server, enable them by setting the "
                              "configuration target version to a value "
                              "greater than or equal to '6.1' "
                              "(--target-version command line argument).")

        if only_large_comms and \
            version.parse(self.target_version) < version.parse("6.1"):
            comms = only_large_comms
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "large_communities",
                "The communit{y_ies} '{names}' ha{s_ve} been configured to be "
                "implemented using only the large communit{y_ies} "
                "'{comms}'; large communities are not supported by "
                "OpenBGPD previous to OpenBSD 6.1 so the function{s1} "
                "{it_they} {is_are} used for will be not available. "
                "{large_comms_advice}".format(
                    y_ies="y" if len(comms) == 1 else "ies",
                    names=", ".join([_[0] for _ in comms]),
                    s_ve="s" if len(comms) == 1 else "ve",
                    comms=", ".join([_[1] for _ in comms]),
                    s1="" if len(comms) == 1 else "s",
                    it_they="it" if len(comms) == 1 else "they",
                    is_are="is" if len(comms) == 1 else "are",
                    large_comms_advice=large_comms_advice
                )
            ):
                res = False

        if large_comms_used and \
            version.parse(self.target_version) < version.parse("6.1"):
            comms = large_comms_used
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "large_communities",
                "The communit{y_ies} '{names}' ha{s_ve} been configured to be "
                "implemented using also the large communit{y_ies} "
                "'{comms}'; large communities are not supported by "
                "OpenBGPD previous to OpenBSD 6.1 so the function{s1} "
                "{it_they} {is_are} used for will be available via "
                "standard/extended communities only. "
                "{large_comms_advice}".format(
                    y_ies="y" if len(comms) == 1 else "ies",
                    names=", ".join([_[0] for _ in comms]),
                    s_ve="s" if len(comms) == 1 else "ve",
                    comms=", ".join([_[1] for _ in comms]),
                    s1="" if len(comms) == 1 else "s",
                    it_they="it" if len(comms) == 1 else "they",
                    is_are="is" if len(comms) == 1 else "are",
                    large_comms_advice=large_comms_advice
                )
            ):
                res = False

        if peer_as_ext_comms:
            comms = peer_as_ext_comms
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "extended_communities",
                "The peer-ASN-specific communit{y_ies} '{names}' "
                "ha{s_ve} been configured to be implemented using the "
                "extended communit{y_ies} '{comms}'; please be aware that "
                "peer-ASN-specific extended communities are not scrubbed "
                "from routes that leave OpenBGPD route servers and they are "
                "propagated to the route server clients.".format(
                    y_ies="y" if len(comms) == 1 else "ies",
                    names=", ".join([_[0] for _ in comms]),
                    s_ve="s" if len(comms) == 1 else "ve",
                    comms=", ".join([_[1] for _ in comms])
                )
            ):
                res = False

        try:
            self.cfg_general.check_overlapping_communities(
                allow_private_asns=False)
        except ConfigError as e:
            res = False
            logging.error("{}OpenBGPD doesn't allow to delete BGP "
                          "communities using ranges of values, but only "
                          "using the wildcard ('*'), so also "
                          "outbound communities whose last part contain "
                          "private ASNs collide with inbound communities "
                          "that use the 'peer_as' macro.".format(
                              str(e) + " " if str(e) else ""
                            ))

        if (self.cfg_general["graceful_shutdown"]["enabled"] or \
            self.perform_graceful_shutdown) and \
            version.parse(self.target_version) <= version.parse("6.1"):
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "graceful_shutdown",
                "GRACEFUL_SHUTDOWN BGP community is not implemented "
                "on OpenBGPD versions prior to 6.2. By marking this issue "
                "as ignored the graceful shutdown option will not be "
                "considered and the feature will be not included into the "
                "configuration. "
                "If the release running on the route server is 6.2 or later "
                "please consider to enable this feature by "
                "setting the configuration target version to a value "
                "greater than or equal to '6.2' (--target-version command "
                "line argument)."
            ):
                res = False

        internal_communities_collistion = []
        for comm_name in ConfigParserGeneral.COMMUNITIES_SCHEMA:
            comm = self.cfg_general["communities"][comm_name]
            if comm["ext"] and comm["ext"].startswith("ro:65535:"):
                internal_communities_collistion.append(comm_name)
        if internal_communities_collistion:
            if not self.process_bgpspeaker_specific_compatibility_issue(
                "internal_communities",
                "The Extended BGP communities in the range ro:65535:* "
                "are reserved for internal purposes. "
                "A collision has been detected with the following "
                "communit{y_ies}: {comms}".format(
                    y_ies="y" if len(internal_communities_collistion) == 1 else "ies",
                    comms=", ".join(internal_communities_collistion)
                )
            ):
                res = False

        # Kept for backward compatibility
        use_rpki_roas_as_route_objects_cfg = \
            self.cfg_general["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]
        if use_rpki_roas_as_route_objects_cfg["enabled"]:
            if self.cfg_general["rpki_roas"]["source"] != "ripe-rpki-validator-cache":
                if not self.process_bgpspeaker_specific_compatibility_issue(
                    "rpki_roas_as_route_objects_source",
                    "For OpenBGPD only the 'ripe-rpki-validator-cache' "
                    "value is allowed for the 'rpki_roas.source' option."
                ):
                    res = False

        if self.cfg_general.rpki_roas_needed:
            if self.cfg_general["rpki_roas"]["source"] != "ripe-rpki-validator-cache":
                if not self.process_bgpspeaker_specific_compatibility_issue(
                    "rpki_roas_source",
                    "For OpenBGPD only the 'ripe-rpki-validator-cache' "
                    "value is allowed for the 'rpki_roas.source' option."
                ):
                    res = False

        return res

    def enrich_j2_environment(self, env):

        def convert_ext_comm(s):
            parts = s.split(":")
            return "{} {}:{}".format(
                parts[0], parts[1], parts[2]
            )

        def at_least_one_client_uses_tag_reject_policy():
            for client in self.cfg_clients.cfg["clients"]:
                policy = client["cfg"]["filtering"]["reject_policy"]["policy"]
                if policy == "tag":
                    return True
            return False

        def community_is_set(comm):
            if not comm:
                return False
            # OpenBGPD <= 6.0 does not implement large BGP communities,
            # so only standard and extended ones are considered.
            if version.parse(self.target_version) < version.parse("6.1"):
                if not comm["std"] and not comm["ext"]:
                    return False
            else:
                if not comm["std"] and not comm["ext"] and not comm["lrg"]:
                    return False
            return True

        def aggregated_roas_covered_space():
            prefixes = []
            for pref_len in self.rpki_roas:
                for roa in self.rpki_roas[pref_len]:
                    prefixes.append(roa["prefix"])
            return aggregate(prefixes)

        env.filters["convert_ext_comm"] = convert_ext_comm
        env.filters["community_is_set"] = community_is_set
        self.data["at_least_one_client_uses_tag_reject_policy"] = \
            at_least_one_client_uses_tag_reject_policy()
        self.data["rpki_roas_covered_space"] = aggregated_roas_covered_space()

class TemplateContextDumper(ConfigBuilder):

    def enrich_j2_environment(self, env):

        def to_yaml(obj):
            return yaml.safe_dump(obj, default_flow_style=False)

        def parse_irrdb_info(irrdb_info):
            lst = []
            for bundle_id in irrdb_info:
                bundle = irrdb_info[bundle_id]
                lst.append(bundle.to_dict())
            return lst

        def parse_generic_irr_whois_records(records):
            res = {}
            for origin_asn in records:
                res[origin_asn] = list(records[origin_asn].prefixes)
            return res

        env.filters["to_yaml"] = to_yaml
        env.filters["parse_irrdb_info"] = parse_irrdb_info
        env.filters["parse_generic_irr_whois_records"] = parse_generic_irr_whois_records



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import sys

from .base import ARouteServerCommand
from ..builder import ConfigBuilder, BIRDConfigBuilder, \
                      OpenBGPDConfigBuilder, TemplateContextDumper
from ..config.program import program_config
from ..errors import ARouteServerError, TemplateRenderingError

class TemplateRenderingCommands(ARouteServerCommand):

    NEEDS_CONFIG = True

    BUILDER_CLASS = None

    @classmethod
    def add_arguments(cls, parser):
        super(TemplateRenderingCommands, cls).add_arguments(parser)

        parser.add_argument(
            "-o", "--output",
            type=argparse.FileType('w'),
            help="Output file. Default: stdout.",
            default=sys.stdout,
            dest="output_file")

        parser.add_argument(
            "--test-only",
            action="store_true",
            help="Only verify the input configuration files (general.yml, "
                 "clients.yml and so on), do not produce any output "
                 "configuration.",
            dest="test_only")

        parser.add_argument(
            "--ignore-issues",
            nargs="+",
            help="Ignore compatibility issues identified by the IDs "
                 "provided here.",
            metavar="ISSUE_ID",
            dest="ignore_errors")

        group = parser.add_argument_group(
            title="Route server configuration",
            description="The following arguments override those provided "
                        "in the program's configuration file."
        )

        group.add_argument(
            "--general",
            help="General route server configuration file.",
            metavar="FILE",
            dest="cfg_general")

        group.add_argument(
            "--clients",
            help="Route server clients configuration file.",
            metavar="FILE",
            dest="cfg_clients")

        group.add_argument(
            "--bogons",
            help="Bogons configuration file.",
            metavar="FILE",
            dest="cfg_bogons")

        group = parser.add_argument_group(
            title="Rendering",
            description="The following arguments override those provided "
                        "in the program's configuration file."
        )

        group.add_argument(
            "--templates-dir",
            help="Directory where Jinja2 files are stored. This is the "
                 "directory where the \"html\" directory and other BGP "
                 "speaker specific directories (\"bird\") can be found.",
            metavar="DIR",
            dest="templates_dir")

        group.add_argument(
            "--template-file-name",
            help="Main Jinja2 template file name.",
            metavar="NAME",
            dest="template_name")

        group.add_argument(
            "--ip-ver",
            help="IP version. "
                 "Default: both IPv4 and IPv6",
            default=None,
            choices=[4, 6],
            type=int,
            dest="ip_ver")

        group.add_argument(
            "--perform-graceful-shutdown",
            help="When set, the configuration built by the program "
                 "includes an outbound policy which is applied to "
                 "BGP sessions toward the clients and which adds "
                 "the GRACEFUL_SHUTDOWN BGP community (65535:0) to "
                 "all the routes that the route server announces to "
                 "them. This can be useful to perform a graceful "
                 "shutdown of the route server itself. "
                 "This argument works regardless of the setting of "
                 "graceful_shutdown.enabled option in general.yml file. "
                 "Please note: the configuration generated when this "
                 "flag is set should be used only temporarly and "
                 "must be replaced with the production configuration "
                 "before the route server is reloaded.",
            action="store_true",
            dest="perform_graceful_shutdown")

    def _get_template_sub_dir(self):
        raise NotImplementedError()

    def _set_cfg_builder_params(self):
        pass

    def run(self):
        tpl_all_right = program_config.verify_templates() == []
        if not tpl_all_right:
            logging.warning("One or more templates are not aligned "
                            "with those used by the current version "
                            "of the program. "
                            "Run 'arouteserver verify-templates' for "
                            "more information.")

        # Config builder setup
        self.cfg_builder_params = {
            "cfg_general": program_config.get("cfg_general"),
            "cfg_clients": program_config.get("cfg_clients"),
            "cfg_bogons": program_config.get("cfg_bogons"),
            "cache_dir": program_config.get_dir("cache_dir"),
            "cache_expiry": program_config.get("cache_expiry"),
            "bgpq3_path": program_config.get("bgpq3_path"),
            "bgpq3_host": program_config.get("bgpq3_host"),
            "bgpq3_sources": program_config.get("bgpq3_sources"),
            "rtt_getter_path": program_config.get("rtt_getter_path"),
            "template_dir": program_config.get_dir("templates_dir"),
            "template_name": program_config.get("template_name"),
            "ip_ver": self.args.ip_ver,
            "perform_graceful_shutdown": self.args.perform_graceful_shutdown,
            "threads": program_config.get("threads"),
            "ignore_errors": self.args.ignore_errors,
        }
        self._set_cfg_builder_params()

        builder_class = self.BUILDER_CLASS

        template_sub_dir = self._get_template_sub_dir()
        if template_sub_dir:
            self.cfg_builder_params["template_dir"] = os.path.join(
                self.cfg_builder_params["template_dir"], template_sub_dir
            )

        try:
            builder = builder_class(**self.cfg_builder_params)
            if not self.args.test_only:
                builder.render_template(output_file=self.args.output_file)
        except TemplateRenderingError as e:
            if tpl_all_right:
                raise
            e.templates_not_aligned = True
            raise e

        return True

class ConfigRenderingCommand(TemplateRenderingCommands):

    @classmethod
    def add_arguments(cls, parser):
        super(ConfigRenderingCommand, cls).add_arguments(parser)

        URL = ("https://arouteserver.readthedocs.io/en/latest/CONFIG.html"
               "#site-specific-custom-config")

        cls.customization_group = parser.add_argument_group(
            title="Local customizations and site-specific configurations",
            description="The following arguments can be used to enable "
                        "the inclusion of .local files into the configuration "
                        "generated by the program, to allow the integration "
                        "of custom functionalities implemented locally to the "
                        "route server. Details can be found at this URL: "
                        "{}".format(URL)
        )

        cls.customization_group.add_argument(
            "--local-files-dir",
            help="The directory where .local files are located, from the "
                 "route server's perspective.",
            default=cls.BUILDER_CLASS.LOCAL_FILES_BASE_DIR,
            dest="local_files_dir"
        )

        cls.customization_group.add_argument(
            "--use-local-files",
            help="Enable the inclusion of .local files into the configuration "
                 "generated by the program. "
                 "The list of available .local files IDs follows: {}".format(
                     ", ".join(cls.BUILDER_CLASS.LOCAL_FILES_IDS)
                 ),
            nargs="*",
            choices=cls.BUILDER_CLASS.LOCAL_FILES_IDS,
            metavar="FILE_ID",
            dest="local_files")

        parser.add_argument(
            "--target-version",
            help="The version of the BGP daemon for which the configuration "
                 "will be generated. Default for {}: {}".format(
                     cls.COMMAND_NAME,
                     cls.BUILDER_CLASS.DEFAULT_VERSION
                 ),
            dest="target_version",
            choices=cls.BUILDER_CLASS.AVAILABLE_VERSION,
            default=cls.BUILDER_CLASS.DEFAULT_VERSION)

    def _set_cfg_builder_params(self):
        super(ConfigRenderingCommand, self)._set_cfg_builder_params()

        self.cfg_builder_params["local_files_dir"] = self.args.local_files_dir
        self.cfg_builder_params["local_files"] = self.args.local_files
        self.cfg_builder_params["target_version"] = self.args.target_version

class BuildCommand(TemplateRenderingCommands):

    COMMAND_NAME = "build"
    COMMAND_HELP = ("Build route server configuration. DEPRECATED! "
                    "Please use 'bird' or 'openbgpd' commands.")

    @classmethod
    def add_arguments(cls, parser):
        super(BuildCommand, cls).add_arguments(parser)

        parser.add_argument(
            "--speaker",
            help="The BGP speaker target implementation for "
                 "the configuration that will be built.",
            dest="speaker",
            choices=["BIRD"],
            default="BIRD")

    def run(self):
        raise ARouteServerError(
            "The 'build' command has been replaced by "
            "BGP-speaker-specific commands: 'bird' and 'openbgpd'."
        )

class BIRDCommand(ConfigRenderingCommand):

    COMMAND_NAME = "bird"
    COMMAND_HELP = "Build route server configuration for BIRD."

    BUILDER_CLASS = BIRDConfigBuilder

    def _get_template_sub_dir(self):
        return "bird"

    @classmethod
    def add_arguments(cls, parser):
        super(BIRDCommand, cls).add_arguments(parser)

        cls.customization_group.add_argument(
            "--use-hooks",
            help="Enable the use of function hooks to add custom "
                 "functionalities. Hooks must be implemented in .local "
                 "files, accordingly to specifications documented at the "
                 "URL shown above."
                 "The list of available hooks follows: {}".format(
                     ", ".join(cls.BUILDER_CLASS.HOOKS)
                 ),
            nargs="*",
            choices=cls.BUILDER_CLASS.HOOKS,
            metavar="HOOK_NAME",
            dest="hooks")

    def _set_cfg_builder_params(self):
        super(BIRDCommand, self)._set_cfg_builder_params()

        self.cfg_builder_params["hooks"] = self.args.hooks

class OpenBGPDCommand(ConfigRenderingCommand):

    COMMAND_NAME = "openbgpd"
    COMMAND_HELP = "Build route server configuration for OpenBGPD."

    BUILDER_CLASS = OpenBGPDConfigBuilder

    def _get_template_sub_dir(self):
        return "openbgpd"

class HTMLCommand(TemplateRenderingCommands):

    COMMAND_NAME = "html"
    COMMAND_HELP = ("Build an HTML descriptive page containing the "
                    "textual representation of the route server "
                    "configuration.")

    BUILDER_CLASS = ConfigBuilder

    def _get_template_sub_dir(self):
        return "html"

class DumpTemplateContextCommand(TemplateRenderingCommands):

    COMMAND_NAME = "template-context"
    COMMAND_HELP = ("Dump the context used to build templates, that is the "
                    "data and the variables that can be consumed within a "
                    "template to build the output configuration.")

    BUILDER_CLASS = TemplateContextDumper

    def _get_template_sub_dir(self):
        return "template-context"



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six
import unittest

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class PathHidingScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    CFG_GENERAL = None

    AS_SET = {
    }
    R_SET = {
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS3",
                cls.DATA["AS3_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS3.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS4",
                cls.DATA["AS4_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS4.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS101",
                cls.DATA["AS101_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS101.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.AS3 = self._get_instance_by_name("AS3")
        self.AS4 = self._get_instance_by_name("AS4")
        self.AS101 = self._get_instance_by_name("AS101")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS3)
        self.session_is_up(self.rs, self.AS4)
        self.session_is_up(self.AS1, self.AS101)
        self.session_is_up(self.AS2, self.AS101)

    def test_030_rs_receives_route_from_AS1_and_AS2(self):
        """{}: rs should receive prefix from both AS1 and AS2"""
        self.receive_route(self.rs, self.DATA["AS101_pref_ok1"], self.AS1,
                           as_path="1 101")
        self.receive_route(self.rs, self.DATA["AS101_pref_ok1"], self.AS2,
                           as_path="2 101 101 101 101")

    def test_031_rs_has_best_toward_AS1(self):
        """{}: rs should have best toward AS1"""
        self.receive_route(self.rs, self.DATA["AS101_pref_ok1"], self.AS1,
                           only_best=True)

    def test_032_AS1_do_not_announce_to_AS3_and_AS4(self):
        """{}: AS1 wants rs to not announce to AS3 and AS4"""
        self.receive_route(self.rs, self.DATA["AS101_pref_ok1"], self.AS1,
                           std_comms=["0:3", "0:4"])
        self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{AS3}}".format(
            self.DATA["AS101_pref_ok1"]), {"AS3": self.AS3})
        self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{AS4}}".format(
            self.DATA["AS101_pref_ok1"]), {"AS4": self.AS4})

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()
        self.test_020_sessions_up()

class PathHidingScenarioBIRD(PathHidingScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    TARGET_VERSION = None
    IP_VER = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     cfg_general=cls.CFG_GENERAL,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bird/bird.conf"
                )
            ]
        )

class PathHidingScenarioBIRD2(PathHidingScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

class PathHidingScenarioOpenBGPD(LiveScenario_TagRejectPolicy, PathHidingScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     cfg_general=cls._get_cfg_general(cls.CFG_GENERAL),
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )

class PathHidingScenarioOpenBGPDPrevious(PathHidingScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.BGP_SPEAKER_VERSION

class PathHidingScenarioOpenBGPDLatest(PathHidingScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.BGP_SPEAKER_VERSION

class PathHidingScenario_MitigationOn(object):

    CFG_GENERAL = "general_on.yml"

    def test_040_AS3_and_AS4_prefix_via_AS2(self):
        """{}: AS3 and AS4 receive prefix with sub-optimal path via AS2"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("Work in progress")

        for inst in (self.AS3, self.AS4):
            self.receive_route(inst, self.DATA["AS101_pref_ok1"], self.rs,
                               as_path="2 101 101 101 101", next_hop=self.AS2,
                               std_comms=[])

    def test_041_AS3_and_AS4_no_prefix_via_AS1(self):
        """{}: AS3 and AS4 don't receive prefix via AS1"""
        for inst in (self.AS3, self.AS4):
            with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS101_pref_ok1"], self.rs,
                                   next_hop=self.AS1)

class PathHidingScenario_MitigationOff(object):

    CFG_GENERAL = "general_off.yml"

    def test_050_AS3_prefix_not_received_by_AS3(self):
        """{}: AS3 does not receive prefix at all"""
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS3, self.DATA["AS101_pref_ok1"])

    def test_051_AS4_receives_prefix_via_AS2_because_of_ADD_PATH(self):
        """{}: AS4 receives the prefix via AS2 because of ADD-PATH"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("ADD-PATH not supported by OpenBGPD")

        self.receive_route(self.AS4, self.DATA["AS101_pref_ok1"], self.rs,
                           as_path="2 101 101 101 101", next_hop=self.AS2,
                           std_comms=[])
